from __future__ import annotations

import itertools
from collections.abc import (
    Callable,
    Generator,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
    ValuesView,
)
from functools import partial
from typing import TYPE_CHECKING, Any, Concatenate, overload

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper

if TYPE_CHECKING:
    from .._results import Option
    from ._main import Iter


class BaseMap[T](IterWrapper[T]):
    def for_each[**P](
        self,
        func: Callable[Concatenate[T, P], Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """Consume the Iterator by applying a function to each element in the iterable.

        Args:
            func (Callable[Concatenate[T, P], Any]): Function to apply to each element.
            *args (P.args): Positional arguments for the function.
            **kwargs (P.kwargs): Keyword arguments for the function.

        Returns:
            None: This is a terminal operation with no return value.

        Is a terminal operation, and is useful for functions that have side effects,
        or when you want to force evaluation of a lazy iterable.

        Examples:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).for_each(lambda x: print(x))
        1
        2
        3

        ```

        """

        def _for_each(data: Iterable[T]) -> None:
            for v in data:
                func(v, *args, **kwargs)

        return self.into(_for_each)

    def map[R](self, func: Callable[[T], R]) -> Iter[R]:
        """Map each element through func and return a Iter of results.

        Args:
            func (Callable[[T], R]): Function to apply to each element.

        Returns:
            Iter[R]: An iterable of transformed elements.
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).map(lambda x: x + 1).into(list)
        [2, 3]

        ```
        """
        return self._lazy(partial(map, func))

    @overload
    def flat_map[U, R](
        self: IterWrapper[Iterable[U]],
        func: Callable[[U], R],
    ) -> Iter[R]: ...
    @overload
    def flat_map[U, R](
        self: IterWrapper[Iterator[U]],
        func: Callable[[U], R],
    ) -> Iter[R]: ...
    @overload
    def flat_map[U, R](
        self: IterWrapper[Sequence[U]],
        func: Callable[[U], R],
    ) -> Iter[R]: ...
    @overload
    def flat_map[U, R](
        self: IterWrapper[list[U]],
        func: Callable[[U], R],
    ) -> Iter[R]: ...
    @overload
    def flat_map[U, R](
        self: IterWrapper[tuple[U, ...]],
        func: Callable[[U], R],
    ) -> Iter[R]: ...
    @overload
    def flat_map[R](self: IterWrapper[range], func: Callable[[int], R]) -> Iter[R]: ...
    def flat_map[U: Iterable[Any], R](
        self: IterWrapper[U],
        func: Callable[[Any], R],
    ) -> Iter[Any]:
        """Map each element through func and flatten the result by one level.

        Args:
            func (Callable[[Any], R]): Function to apply to each element.

        Returns:
            Iter[Any]: An iterable of flattened transformed elements.
        >>> import pyochain as pc
        >>> data = [[1, 2], [3, 4]]
        >>> pc.Iter.from_(data).flat_map(lambda x: x + 10).into(list)
        [11, 12, 13, 14]

        ```
        """

        def _flat_map(data: Iterable[U]) -> map[R]:
            return map(func, itertools.chain.from_iterable(data))

        return self._lazy(_flat_map)

    def map_star[U: Iterable[Any], R](
        self: IterWrapper[U],
        func: Callable[..., R],
    ) -> Iter[R]:
        """Applies a function to each element.where each element is an iterable.

        Unlike `.map()`, which passes each element as a single argument,
        `.starmap()` unpacks each element into positional arguments for the function.

        In short, for each `element` in the sequence, it computes `func(*element)`.

        Args:
            func (Callable[..., R]): Function to apply to unpacked elements.

        Returns:
            Iter[R]: An iterable of results from applying the function to unpacked elements.
        >>> import pyochain as pc
        >>> def make_sku(color, size):
        ...     return f"{color}-{size}"
        >>> data = pc.Seq(["blue", "red"])
        >>> data.iter().product(["S", "M"]).map_star(make_sku).into(list)
        ['blue-S', 'blue-M', 'red-S', 'red-M']

        ```
        This is equivalent to:
        ```python
        >>> data.iter().product(["S", "M"]).map(lambda x: make_sku(*x)).into(list)
        ['blue-S', 'blue-M', 'red-S', 'red-M']

        ```

        - Use map_star when the performance matters (it is faster).
        - Use map with unpacking when readability matters (the types can be inferred).

        """
        return self._lazy(partial(itertools.starmap, func))

    def map_if[R](
        self,
        predicate: Callable[[T], bool],
        func: Callable[[T], R],
        func_else: Callable[[T], R] | None = None,
    ) -> Iter[R]:
        """Evaluate each item with `func` or `func_else`, based on the `predicate` result.

        If the result is equivalent to True, transform the item with func and yield it.

        Otherwise, the presence or absence of `func_else` determines the strategy:
            - If `func_else` is provided, transform the item with it and yield it.
            - If `func_else` is not provided, skip the item (do not yield it).


        Args:
            predicate (Callable[[T], bool]): Function to evaluate each item.
            func (Callable[[T], R]): Function to apply if predicate is True.
            func_else (Callable[[T], R] | None): Function to apply if predicate is False.

        Returns:
            Iter[R]: An iterable of transformed and filtered items.

        ```python
        >>> import pyochain as pc
        >>> from math import sqrt
        >>> data = pc.Seq(range(-5, 5))
        >>> data.iter().map_if(lambda x: x > 3, lambda x: f"{x} is toobig").into(list)
        ['4 is toobig']
        >>> data.iter().map_if(
        ...     lambda x: x >= 0,
        ...     lambda x: f"{sqrt(x):.2f}",
        ...     lambda x: None,
        ... ).collect()
        Seq([None, None, None, None, None, '0.00', '1.00', '1.41', '1.73', '2.00'])

        ```

        """

        def _map_if(data: Iterable[T]) -> Generator[R, None, None]:
            match func_else:
                case None:
                    return (func(item) for item in data if predicate(item))
                case _:
                    return (
                        func(item) if predicate(item) else func_else(item)
                        for item in data
                    )

        return self._lazy(_map_if)

    def map_except[R](
        self,
        func: Callable[[T], R],
        *exceptions: type[BaseException],
    ) -> Iter[R]:
        """Transform each item from iterable with function and yield the result, unless function raises one of the specified exceptions.

        Args:
            func (Callable[[T], R]): Function to apply to each item.
            *exceptions (type[BaseException]): Exceptions to catch and ignore.

        Returns:
            Iter[R]: An iterable of transformed items that don't raise specified exceptions.

        The function is called to transform each item in iterable

        If an exception other than one given by exceptions is raised by function, it is raised like normal.
        ```python
        >>> import pyochain as pc
        >>> iterable = ["1", "2", "three", "4", None]
        >>> pc.Iter.from_(iterable).map_except(int, ValueError, TypeError).into(list)
        [1, 2, 4]

        ```

        """

        def _map_except(data: Iterable[T]) -> Iterator[R]:
            return mit.map_except(func, data, *exceptions)

        return self._lazy(_map_except)

    def repeat(
        self,
        n: int,
        factory: Callable[[Iterable[T]], Sequence[T]] = tuple,
    ) -> Iter[Iterable[T]]:
        """Repeat the entire iterable n times (as elements).

        Args:
            n (int): Number of repetitions.
            factory (Callable[[Iterable[T]], Sequence[T]]): Factory to create the repeated Sequence (default: tuple).

        Returns:
            Iter[Iterable[T]]: An iterable of repeated sequences.
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).repeat(2).collect()
        Seq([(1, 2), (1, 2)])
        >>> pc.Iter.from_([1, 2]).repeat(3, list).collect()
        Seq([[1, 2], [1, 2], [1, 2]])

        ```
        """

        def _repeat(data: Iterable[T]) -> Iterator[Iterable[T]]:
            return itertools.repeat(factory(data), n)

        return self._lazy(_repeat)

    @overload
    def repeat_last(self, default: T) -> Iter[T]: ...
    @overload
    def repeat_last[U](self, default: U) -> Iter[T | U]: ...
    def repeat_last[U](self, default: U = None) -> Iter[T | U]:
        """After the iterable is exhausted, keep yielding its last element.

        **Warning** ⚠️
            This creates an infinite iterator.
            Be sure to use `Iter.take()` or `Iter.slice()` to limit the number of items taken.

        Args:
            default (U): Value to yield if the iterable is empty.

        Returns:
            Iter[T | U]: An iterable that yields the last element repeatedly, or default if empty
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(range(3)).repeat_last().take(5).into(list)
        [0, 1, 2, 2, 2]

        If the iterable is empty, yield default forever:
        ```python
        >>> pc.Iter.from_(range(0)).repeat_last(42).take(5).into(list)
        [42, 42, 42, 42, 42]

        ```

        """
        return self._lazy(mit.repeat_last, default)

    def ichunked(self, n: int) -> Iter[Iterator[T]]:
        """Break *iterable* into sub-iterables with *n* elements each.

        Args:
            n (int): Number of elements in each chunk.

        Returns:
            Iter[Iterator[T]]: An iterable of iterators, each yielding n elements.
        If the sub-iterables are read in order, the elements of *iterable*
        won't be stored in memory.

        If they are read out of order, :func:`itertools.tee` is used to cache
        elements as necessary.
        ```python
        >>> import pyochain as pc
        >>> all_chunks = pc.Iter.from_count().ichunked(4)
        >>> c_1, c_2, c_3 = all_chunks.next(), all_chunks.next(), all_chunks.next()
        >>> list(c_2)  # c_1's elements have been cached; c_3's haven't been
        [4, 5, 6, 7]
        >>> list(c_1)
        [0, 1, 2, 3]
        >>> list(c_3)
        [8, 9, 10, 11]

        ```

        """
        return self._lazy(mit.ichunked, n)

    @overload
    def flatten[U](self: IterWrapper[Generator[U, None, None]]) -> Iter[U]: ...
    @overload
    def flatten[U](self: IterWrapper[ValuesView[U]]) -> Iter[U]: ...
    @overload
    def flatten[U](self: IterWrapper[Iterable[U]]) -> Iter[U]: ...
    @overload
    def flatten[U](self: IterWrapper[Iterator[U]]) -> Iter[U]: ...
    @overload
    def flatten[U](self: IterWrapper[Sequence[U]]) -> Iter[U]: ...
    @overload
    def flatten[U](self: IterWrapper[list[U]]) -> Iter[U]: ...
    @overload
    def flatten[U](self: IterWrapper[tuple[U, ...]]) -> Iter[U]: ...
    @overload
    def flatten(self: IterWrapper[range]) -> Iter[int]: ...
    def flatten[U: Iterable[Any]](self: IterWrapper[U]) -> Iter[Any]:
        """Flatten one level of nesting and return a new Iterable wrapper.

        This is a shortcut for `.apply(itertools.chain.from_iterable)`.

        Returns:
            Iter[Any]: An iterable of flattened elements.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([[1, 2], [3]]).flatten().into(list)
        [1, 2, 3]

        ```
        """
        return self._lazy(itertools.chain.from_iterable)

    def pluck[U: Mapping[Any, Any]](
        self: IterWrapper[U],
        *keys: str | int,
    ) -> Iter[Any]:
        """Get an element from each item in a sequence using a nested key path.

        Args:
            *keys (str | int): Nested keys to extract values.

        Returns:
            Iter[Any]: An iterable of extracted values.
        ```python
        >>> import pyochain as pc
        >>> data = pc.Seq(
        ...     [
        ...         {"id": 1, "info": {"name": "Alice", "age": 30}},
        ...         {"id": 2, "info": {"name": "Bob", "age": 25}},
        ...     ]
        ... )
        >>> data.iter().pluck("info").into(list)
        [{'name': 'Alice', 'age': 30}, {'name': 'Bob', 'age': 25}]
        >>> data.iter().pluck("info", "name").into(list)
        ['Alice', 'Bob']

        ```
        Example: get the maximum age along with the corresponding id)
        ```python
        >>> data.iter().pluck("info", "age").zip(
        ...     data.iter().pluck("id").into(list)
        ... ).max()
        (30, 1)

        ```

        """
        getter = partial(cz.dicttoolz.get_in, keys)
        return self._lazy(partial(map, getter))

    def scan[U](self, state: U, func: Callable[[U, T], Option[U]]) -> Iter[U]:
        """Transform elements by sharing state between iterations.

        `scan` takes two arguments:
            - an initial value which seeds the internal state
            - a closure with two arguments

        The first being a reference to the internal state and the second an iterator element.

        The closure can assign to the internal state to share state between iterations.

        On iteration, the closure will be applied to each element of the iterator and the return value from the closure, an Option, is returned by the next method.

        Thus the closure can return Some(value) to yield value, or None to end the iteration.

        Args:
            state (U): Initial state.
            func (Callable[[U, T], Option[U]]): Function that takes the current state and an item, and returns an Option.

        Returns:
            Iter[U]: An iterable of the yielded values.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def accumulate_until_limit(state: int, item: int) -> pc.Option[int]:
        ...     new_state = state + item
        ...     match new_state:
        ...         case _ if new_state <= 10:
        ...             return pc.Some(new_state)
        ...         case _:
        ...             return pc.NONE
        >>> pc.Seq([1, 2, 3, 4, 5]).iter().scan(0, accumulate_until_limit).collect()
        Seq([1, 3, 6, 10])

        ```

        """

        def gen(data: Iterable[T]) -> Iterator[U]:
            current: U = state
            for item in data:
                res = func(current, item)
                if res.is_none():
                    break
                current = res.unwrap()
                yield res.unwrap()

        return self._lazy(gen)



from __future__ import annotations

import itertools
from collections.abc import Callable, Iterable, Iterator
from enum import StrEnum, auto
from functools import partial
from typing import TYPE_CHECKING, Any, Literal, overload

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper

if TYPE_CHECKING:
    from ._main import Iter


class Position(StrEnum):
    """Enum representing the position of an element in an iterable.

    Attributes:
        FIRST: The element is the first in the iterable.
        MIDDLE: The element is neither the first nor the last in the iterable.
        LAST: The element is the last in the iterable.
        ONLY: The element is the only one in the iterable.
    """

    FIRST = auto()
    MIDDLE = auto()
    LAST = auto()
    ONLY = auto()


class BaseTuples[T](IterWrapper[T]):
    def enumerate(self) -> Iter[tuple[int, T]]:
        """Return a Iter of (index, value) pairs.

        Returns:
            Iter[tuple[int, T]]: An iterable of (index, value) pairs.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(["a", "b"]).enumerate().into(list)
        [(0, 'a'), (1, 'b')]

        ```

        """
        return self._lazy(enumerate)

    @overload
    def combinations(self, r: Literal[2]) -> Iter[tuple[T, T]]: ...
    @overload
    def combinations(self, r: Literal[3]) -> Iter[tuple[T, T, T]]: ...
    @overload
    def combinations(self, r: Literal[4]) -> Iter[tuple[T, T, T, T]]: ...
    @overload
    def combinations(self, r: Literal[5]) -> Iter[tuple[T, T, T, T, T]]: ...
    def combinations(self, r: int) -> Iter[tuple[T, ...]]:
        """Return all combinations of length r.

        Args:
            r (int): Length of each combination.

        Returns:
            Iter[tuple[T, ...]]: An iterable of combinations.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Seq([1, 2, 3]).iter().combinations(2).collect()
        Seq([(1, 2), (1, 3), (2, 3)])

        ```

        """
        return self._lazy(itertools.combinations, r)

    @overload
    def permutations(self, r: Literal[2]) -> Iter[tuple[T, T]]: ...
    @overload
    def permutations(self, r: Literal[3]) -> Iter[tuple[T, T, T]]: ...
    @overload
    def permutations(self, r: Literal[4]) -> Iter[tuple[T, T, T, T]]: ...
    @overload
    def permutations(self, r: Literal[5]) -> Iter[tuple[T, T, T, T, T]]: ...
    def permutations(self, r: int | None = None) -> Iter[tuple[T, ...]]:
        """Return all permutations of length r.

        Args:
            r (int | None): Length of each permutation. Defaults to the length of the iterable.

        Returns:
            Iter[tuple[T, ...]]: An iterable of permutations.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Seq([1, 2, 3]).iter().permutations(2).collect()
        Seq([(1, 2), (1, 3), (2, 1), (2, 3), (3, 1), (3, 2)])

        ```

        """
        return self._lazy(itertools.permutations, r)

    @overload
    def combinations_with_replacement(self, r: Literal[2]) -> Iter[tuple[T, T]]: ...
    @overload
    def combinations_with_replacement(self, r: Literal[3]) -> Iter[tuple[T, T, T]]: ...
    @overload
    def combinations_with_replacement(
        self,
        r: Literal[4],
    ) -> Iter[tuple[T, T, T, T]]: ...
    @overload
    def combinations_with_replacement(
        self,
        r: Literal[5],
    ) -> Iter[tuple[T, T, T, T, T]]: ...
    def combinations_with_replacement(self, r: int) -> Iter[tuple[T, ...]]:
        """Return all combinations with replacement of length r.

        Args:
            r (int): Length of each combination.

        Returns:
            Iter[tuple[T, ...]]: An iterable of combinations with replacement.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).combinations_with_replacement(2).into(list)
        [(1, 1), (1, 2), (1, 3), (2, 2), (2, 3), (3, 3)]

        ```

        """
        return self._lazy(itertools.combinations_with_replacement, r)

    def pairwise(self) -> Iter[tuple[T, T]]:
        """Return an iterator over pairs of consecutive elements.

        Returns:
            Iter[tuple[T, T]]: An iterable of pairs of consecutive elements.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).pairwise().into(list)
        [(1, 2), (2, 3)]

        ```

        """
        return self._lazy(itertools.pairwise)

    @overload
    def map_juxt[R1, R2](
        self,
        func1: Callable[[T], R1],
        func2: Callable[[T], R2],
        /,
    ) -> Iter[tuple[R1, R2]]: ...
    @overload
    def map_juxt[R1, R2, R3](
        self,
        func1: Callable[[T], R1],
        func2: Callable[[T], R2],
        func3: Callable[[T], R3],
        /,
    ) -> Iter[tuple[R1, R2, R3]]: ...
    @overload
    def map_juxt[R1, R2, R3, R4](
        self,
        func1: Callable[[T], R1],
        func2: Callable[[T], R2],
        func3: Callable[[T], R3],
        func4: Callable[[T], R4],
        /,
    ) -> Iter[tuple[R1, R2, R3, R4]]: ...
    def map_juxt(self, *funcs: Callable[[T], object]) -> Iter[tuple[object, ...]]:
        """Apply several functions to each item.

        Returns a new Iter where each item is a tuple of the results of applying each function to the original item.

        Args:
            *funcs (Callable[[T], object]): Functions to apply to each item.

        Returns:
            Iter[tuple[object, ...]]: An iterable of tuples containing the results of each function.
        ```python
        >>> import pyochain as pc
        >>> def is_even(n: int) -> bool:
        ...     return n % 2 == 0
        >>> def is_positive(n: int) -> bool:
        ...     return n > 0
        >>>
        >>> pc.Iter.from_([1, -2, 3]).map_juxt(is_even, is_positive).into(list)
        [(False, True), (True, False), (False, True)]

        ```
        """
        return self._lazy(partial(map, cz.functoolz.juxt(*funcs)))

    def adjacent(
        self,
        predicate: Callable[[T], bool],
        distance: int = 1,
    ) -> Iter[tuple[bool, T]]:
        """Return an iterable over (bool, item) tuples.

        Args:
            predicate (Callable[[T], bool]): Function to determine if an item satisfies the condition.
            distance (int): Number of places to consider as adjacent. Defaults to 1.

        Returns:
            Iter[tuple[bool, T]]: An iterable of (bool, item) tuples.

        The output is a sequence of tuples where the item is drawn from iterable.

        The bool indicates whether that item satisfies the predicate or is adjacent to an item that does.

        For example, to find whether items are adjacent to a 3:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(range(6)).adjacent(lambda x: x == 3).into(list)
        [(False, 0), (False, 1), (True, 2), (True, 3), (True, 4), (False, 5)]

        ```
        Set distance to change what counts as adjacent.
        For example, to find whether items are two places away from a 3:
        ```python
        >>> pc.Iter.from_(range(6)).adjacent(lambda x: x == 3, distance=2).into(list)
        [(False, 0), (True, 1), (True, 2), (True, 3), (True, 4), (True, 5)]

        ```

        This is useful for contextualizing the results of a search function.

        For example, a code comparison tool might want to identify lines that have changed, but also surrounding lines to give the viewer of the diff context.

        The predicate function will only be called once for each item in the iterable.

        See also groupby_transform, which can be used with this function to group ranges of items with the same bool value.

        """
        return self._lazy(partial(mit.adjacent, predicate, distance=distance))

    def classify_unique(self) -> Iter[tuple[T, bool, bool]]:
        """Classify each element in terms of its uniqueness.

        For each element in the input iterable, return a 3-tuple consisting of:

        - The element itself
        - False if the element is equal to the one preceding it in the input, True otherwise (i.e. the equivalent of unique_justseen)
        - False if this element has been seen anywhere in the input before, True otherwise (i.e. the equivalent of unique_everseen)

        This function is analogous to unique_everseen and is subject to the same performance considerations.

        Returns:
            Iter[tuple[T, bool, bool]]: An iterable of (element, is_new, is_unique) tuples.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("otto").classify_unique().into(list)
        ... # doctest: +NORMALIZE_WHITESPACE
        [('o', True,  True),
        ('t', True,  True),
        ('t', False, False),
        ('o', True,  False)]

        ```
        """
        return self._lazy(mit.classify_unique)

    @overload
    def group_by_transform(
        self,
        keyfunc: None = None,
        valuefunc: None = None,
        reducefunc: None = None,
    ) -> Iter[tuple[T, Iterator[T]]]: ...
    @overload
    def group_by_transform[U](
        self,
        keyfunc: Callable[[T], U],
        valuefunc: None,
        reducefunc: None,
    ) -> Iter[tuple[U, Iterator[T]]]: ...
    @overload
    def group_by_transform[V](
        self,
        keyfunc: None,
        valuefunc: Callable[[T], V],
        reducefunc: None,
    ) -> Iter[tuple[T, Iterator[V]]]: ...
    @overload
    def group_by_transform[U, V](
        self,
        keyfunc: Callable[[T], U],
        valuefunc: Callable[[T], V],
        reducefunc: None,
    ) -> Iter[tuple[U, Iterator[V]]]: ...
    @overload
    def group_by_transform[W](
        self,
        keyfunc: None,
        valuefunc: None,
        reducefunc: Callable[[Iterator[T]], W],
    ) -> Iter[tuple[T, W]]: ...
    @overload
    def group_by_transform[U, W](
        self,
        keyfunc: Callable[[T], U],
        valuefunc: None,
        reducefunc: Callable[[Iterator[T]], W],
    ) -> Iter[tuple[U, W]]: ...
    @overload
    def group_by_transform[V, W](
        self,
        keyfunc: None,
        valuefunc: Callable[[T], V],
        reducefunc: Callable[[Iterator[V]], W],
    ) -> Iter[tuple[T, W]]: ...
    @overload
    def group_by_transform[U, V, W](
        self,
        keyfunc: Callable[[T], U],
        valuefunc: Callable[[T], V],
        reducefunc: Callable[[Iterator[V]], W],
    ) -> Iter[tuple[U, W]]: ...
    def group_by_transform[U, V](
        self,
        keyfunc: Callable[[T], U] | None = None,
        valuefunc: Callable[[T], V] | None = None,
        reducefunc: Any = None,
    ) -> Iter[tuple[Any, ...]]:
        """An extension of ``Iter.groupby`` that can apply transformations to the grouped data.

        Args:
            keyfunc (Callable[[T], U] | None): Function to compute the key for grouping. Defaults to None.
            valuefunc (Callable[[T], V] | None): Function to transform individual items after grouping. Defaults to None.
            reducefunc (Any): Function to transform each group of items. Defaults to None.

        Returns:
            Iter[tuple[Any, ...]]: An iterable of transformed groups.

        Example:
        ```python
        >>> import pyochain as pc
        >>> data = pc.Iter.from_("aAAbBBcCC")
        >>> data.group_by_transform(
        ...     lambda k: k.upper(), lambda v: v.lower(), lambda g: "".join(g)
        ... ).into(list)
        [('A', 'aaa'), ('B', 'bbb'), ('C', 'ccc')]

        ```
        Each optional argument defaults to an identity function if not specified.

        group_by_transform is useful when grouping elements of an iterable using a separate iterable as the key.

        To do this, zip the iterables and pass a keyfunc that extracts the first element and a valuefunc that extracts the second element:

        Note that the order of items in the iterable is significant.

        Only adjacent items are grouped together, so if you don't want any duplicate groups, you should sort the iterable by the key function.

        Example:
        ```python
        >>> from operator import itemgetter
        >>> data = pc.Iter.from_([0, 0, 1, 1, 1, 2, 2, 2, 3])
        >>> data.zip("abcdefghi").group_by_transform(itemgetter(0), itemgetter(1)).map(
        ...     lambda kv: (kv[0], "".join(kv[1]))
        ... ).into(list)
        [(0, 'ab'), (1, 'cde'), (2, 'fgh'), (3, 'i')]

        ```

        """

        def _group_by_transform(data: Iterable[T]) -> Iterator[tuple[Any, ...]]:
            return mit.groupby_transform(data, keyfunc, valuefunc, reducefunc)

        return self._lazy(_group_by_transform)

    def with_position(self) -> Iter[tuple[Position, T]]:
        """Return an iterable over (`Position`, `T`) tuples.

        The `Position` indicates whether the item `T` is the first, middle, last, or only element in the iterable.

        Returns:
            Iter[tuple[Position, T]]: An iterable of (`Position`, item) tuples.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Seq(["a", "b", "c"]).iter().with_position().collect()
        Seq([(<Position.FIRST: 'first'>, 'a'), (<Position.MIDDLE: 'middle'>, 'b'), (<Position.LAST: 'last'>, 'c')])
        >>> pc.Seq(["a"]).iter().with_position().collect()
        Seq([(<Position.ONLY: 'only'>, 'a')])

        ```

        """

        def gen(data: Iterable[T]) -> Iterator[tuple[Position, T]]:
            it = iter(data)

            try:
                first = next(it)
            except StopIteration:
                return

            try:
                second = next(it)
            except StopIteration:
                yield (Position.ONLY, first)
                return
            yield (Position.FIRST, first)

            current: T = second
            for nxt in it:
                yield (Position.MIDDLE, current)
                current = nxt
            yield (Position.LAST, current)

        return self._lazy(gen)



from __future__ import annotations

from collections.abc import Callable, Iterable, Mapping
from typing import TYPE_CHECKING, Any, Concatenate

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from .._iter import Iter, Seq
    from ._main import Dict


class IterDict[K, V](MappingWrapper[K, V]):
    def itr[**P, R, U](
        self: MappingWrapper[K, Iterable[U]],
        func: Callable[Concatenate[Iter[U], P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Dict[K, R]:
        """Apply a function to each value after wrapping it in an Iter.

        Args:
            func(Callable[Concatenate[Iter[U], P], R]): Function to apply to each value after wrapping it in an Iter.
            *args(P.args): Positional arguments to pass to the function.
            **kwargs(P.kwargs): Keyword arguments to pass to the function.

        Returns:
            Dict[K, R]: Dict with function results as values.

        Syntactic sugar for `map_values(lambda data: func(Iter(data), *args, **kwargs))`
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "numbers1": [1, 2, 3],
        ...     "numbers2": [4, 5, 6],
        ... }
        >>> pc.Dict(data).itr(lambda v: v.repeat(5).flatten().sum()).inner()
        {'numbers1': 30, 'numbers2': 75}

        ```
        """
        from .._iter import Iter

        def _itr(data: Mapping[K, Iterable[U]]) -> dict[K, R]:
            def _(v: Iterable[U]) -> R:
                return func(Iter(iter(v)), *args, **kwargs)

            return cz.dicttoolz.valmap(_, data)

        return self._new(_itr)

    def iter_keys(self) -> Iter[K]:
        """Return an Iter of the dict's keys.

        Returns:
            Iter[K]: An Iter wrapping the dictionary's keys.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2}).iter_keys().collect()
        Seq([1])

        ```
        """
        from .._iter import Iter

        def _keys(data: dict[K, V]) -> Iter[K]:
            return Iter(iter(data.keys()))

        return self.into(_keys)

    def iter_values(self) -> Iter[V]:
        """Return an Iter of the dict's values.

        Returns:
            Iter[V]: An Iter wrapping the dictionary's values.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2}).iter_values().collect()
        Seq([2])

        ```
        """
        from .._iter import Iter

        def _values(data: dict[K, V]) -> Iter[V]:
            return Iter(iter(data.values()))

        return self.into(_values)

    def iter_items(self) -> Iter[tuple[K, V]]:
        """Return an Iter of the dict's items.

        Returns:
            Iter[tuple[K, V]]: An Iter wrapping the dictionary's (key, value) pairs.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2}).iter_items().collect()
        Seq([(1, 2)])

        ```
        """
        from .._iter import Iter

        def _items(data: dict[K, V]) -> Iter[tuple[K, V]]:
            return Iter(iter(data.items()))

        return self.into(_items)

    def to_arrays(self) -> Seq[list[Any]]:
        """Convert the nested dictionary into a sequence of arrays.

        The sequence represents all paths from root to leaves.

        Returns:
            Seq[list[Any]]: A Seq of arrays representing paths from root to leaves.

        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "a": {"b": 1, "c": 2},
        ...     "d": {"e": {"f": 3}},
        ... }
        >>> pc.Dict(data).to_arrays().inner()
        [['a', 'b', 1], ['a', 'c', 2], ['d', 'e', 'f', 3]]

        ```
        """
        from .._iter import Seq

        def _to_arrays(d: Mapping[Any, Any]) -> list[list[Any]]:
            match d:
                case Mapping():
                    arr: list[Any] = []
                    for k, v in d.items():
                        arr.extend([[k, *el] for el in _to_arrays(v)])
                    return arr

                case _:
                    return [[d]]

        return Seq(self.into(_to_arrays))



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Concatenate, Never, TypeIs

from .._core import CommonBase
from ._option import Option, OptionUnwrapError
from ._result import Result, ResultUnwrapError


@dataclass(slots=True)
class Ok[T, E](Result[T, E]):
    """Represents a successful value.

    Attributes:
        value (T): The contained successful value.
    """

    value: T

    def is_ok(self) -> TypeIs[Ok[T, E]]:  # type: ignore[misc]
        return True

    def is_err(self) -> TypeIs[Err[T, E]]:  # type: ignore[misc]
        return False

    def unwrap(self) -> T:
        return self.value

    def unwrap_err(self) -> Never:
        msg = "called `unwrap_err` on Ok"
        raise ResultUnwrapError(msg)


@dataclass(slots=True)
class Err[T, E](Result[T, E]):
    """Represents an error value.

    Attributes:
        error (E): The contained error value.
    """

    error: E

    def is_ok(self) -> TypeIs[Ok[T, E]]:  # type: ignore[misc]
        return False

    def is_err(self) -> TypeIs[Err[T, E]]:  # type: ignore[misc]
        return True

    def unwrap(self) -> Never:
        msg = f"called `unwrap` on Err: {self.error!r}"
        raise ResultUnwrapError(msg)

    def unwrap_err(self) -> E:
        return self.error


@dataclass(slots=True)
class Some[T](Option[T]):
    """Option variant representing the presence of a value.

    Attributes:
        value (T): The contained value.

    Example:
    ```python
    >>> import pyochain as pc
    >>> pc.Some(42)
    Some(value=42)

    ```

    """

    value: T

    def is_some(self) -> TypeIs[Some[T]]:  # type: ignore[misc]
        return True

    def is_none(self) -> TypeIs[NoneOption]:  # type: ignore[misc]
        return False

    def unwrap(self) -> T:
        return self.value


@dataclass(slots=True)
class NoneOption(Option[Any]):
    """Option variant representing the absence of a value."""

    def __repr__(self) -> str:
        return "NONE"

    def is_some(self) -> TypeIs[Some[Any]]:  # type: ignore[misc]
        return False

    def is_none(self) -> TypeIs[NoneOption]:  # type: ignore[misc]
        return True

    def unwrap(self) -> Never:
        msg = "called `unwrap` on a `None`"
        raise OptionUnwrapError(msg)


NONE: Option[Any] = NoneOption()
"""Singleton instance representing the absence of a value."""


class Wrapper[T](CommonBase[T]):
    """A generic Wrapper for any type.

    The pipe into method is implemented to return a Wrapper of the result type.

    This class is intended for use with other types/implementations that do not support the fluent/functional style.

    This allow the use of a consistent code style across the code base.

    """

    def apply[**P, R](
        self,
        func: Callable[Concatenate[T, P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Wrapper[R]:
        return Wrapper(self.into(func, *args, **kwargs))



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Iterator, Sequence
from typing import TYPE_CHECKING, Any, Concatenate, Self
from warnings import deprecated

from ._format import dict_repr

if TYPE_CHECKING:
    from .._dict import Dict
    from .._iter import Iter, Seq


class Pipeable:
    def pipe[**P, R](
        self,
        func: Callable[Concatenate[Self, P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        """Pipe the instance in the function and return the result.

        Accept additional arguments to pass to the function.

        This is a core method, shared by all pyochain wrappers, that allows chaining operations in a functional style.

        Args:
            func (Callable[Concatenate[Self, P], R]): Function to apply to the instance.
            *args (P.args): Positional arguments to pass to the function.
            **kwargs (P.kwargs): Keyword arguments to pass to the function.

        Returns:
            R: The result of the function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def maybe_sum(data: pc.Seq[int]) -> pc.Option[int]:
        ...     match data.count():
        ...         case 0:
        ...             return pc.NONE
        ...         case _:
        ...             return pc.Some(data.sum())
        >>>
        >>> pc.Seq(range(5)).pipe(maybe_sum).unwrap()
        10
        """
        return func(self, *args, **kwargs)


class CommonBase[T](ABC, Pipeable):
    """Base class for all wrappers.

    You can subclass this to create your own wrapper types.
    The pipe unwrap method must be implemented to allow piping functions that transform the underlying data type, whilst retaining the wrapper.

    Args:
        data (T): The underlying data to wrap.
    """

    _inner: T

    __slots__ = ("_inner",)

    def __init__(self, data: T) -> None:
        self._inner = data

    @abstractmethod
    def apply[**P](
        self,
        func: Callable[Concatenate[T, P], Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> CommonBase[Any]:
        raise NotImplementedError

    def println(self, *, pretty: bool = True) -> Self:
        """Print the underlying data and return self for chaining.

        Useful for debugging, simply insert `.println()` in the chain,
        and then removing it will not affect the rest of the chain.

        Args:
            pretty (bool): Whether to pretty print the data. Defaults to True.

        Returns:
            Self: The instance itself for chaining.
        """
        from pprint import pprint

        if pretty:
            self.into(pprint, sort_dicts=False)
        else:
            self.into(print)
        return self

    def inner(self) -> T:
        """Get the underlying data.

        This is a terminal operation that ends the chain.

        Returns:
            T: The underlying data.
        """
        return self._inner

    @deprecated("Use .inner() instead")
    def unwrap(self) -> T:
        """Deprecated: Use `inner()` instead.

        Returns:
            T: The underlying data.
        """
        return self.inner()

    def into[**P, R](
        self,
        func: Callable[Concatenate[T, P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        """Pass the *unwrapped* underlying data into a function.

        The result is not wrapped.

        This is a core functionality that allows ending the chain whilst keeping the code style consistent.

        Args:
            func (Callable[Concatenate[T, P], R]): Function to apply to the underlying data.
            *args (P.args): Positional arguments to pass to the function.
            **kwargs (P.kwargs): Keyword arguments to pass to the function.

        Returns:
            R: The result of the function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Seq(range(5)).into(list)
        [0, 1, 2, 3, 4]

        ```
        """
        return func(self.inner(), *args, **kwargs)

    def equals_to(self, other: Self | T) -> bool:
        """Check if two records are equal based on their data.

        Args:
            other (Self | T): Another instance or corresponding underlying data to compare against.

        Returns:
            bool: True if the underlying data are equal, False otherwise.

        Example:
        ```python
        >>> import pyochain as pc
        >>> d1 = pc.Dict({"a": 1, "b": 2})
        >>> d2 = pc.Dict({"a": 1, "b": 2})
        >>> d3 = pc.Dict({"a": 1, "b": 3})
        >>> d1.equals_to(d2)
        True
        >>> d1.equals_to(d3)
        False

        ```
        """
        other_data = other.inner() if isinstance(other, self.__class__) else other
        return self.inner() == other_data


class IterWrapper[T](CommonBase[Iterable[T]]):
    _inner: Iterable[T]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.inner().__repr__()})"

    def _eager[**P, U](
        self,
        factory: Callable[Concatenate[Iterable[T], P], Sequence[U]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Seq[U]:
        from .._iter import Seq

        def _(data: Iterable[T]) -> Seq[U]:
            return Seq(factory(data, *args, **kwargs))

        return self.into(_)

    def _lazy[**P, U](
        self,
        factory: Callable[Concatenate[Iterable[T], P], Iterator[U]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iter[U]:
        from .._iter import Iter

        def _(data: Iterable[T]) -> Iter[U]:
            return Iter(factory(data, *args, **kwargs))

        return self.into(_)


class MappingWrapper[K, V](CommonBase[dict[K, V]]):
    _inner: dict[K, V]

    def __repr__(self) -> str:
        return f"{self.into(dict_repr)}"

    def _new[KU, VU](self, func: Callable[[dict[K, V]], dict[KU, VU]]) -> Dict[KU, VU]:
        from .._dict import Dict

        return Dict(func(self.inner()))

    def apply[**P, KU, VU](
        self,
        func: Callable[Concatenate[dict[K, V], P], dict[KU, VU]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Dict[KU, VU]:
        """Apply a function to the underlying dict and return a Dict of the result.

        Allow to pass user defined functions that transform the dict while retaining the Dict wrapper.

        Args:
            func (Callable[Concatenate[dict[K, V], P], dict[KU, VU]]): Function to apply to the underlying dict.
            *args (P.args): Positional arguments to pass to the function.
            **kwargs (P.kwargs): Keyword arguments to pass to the function.

        Returns:
            Dict[KU, VU]: A new Dict instance containing the result of the function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def invert_dict(d: dict[K, V]) -> dict[V, K]:
        ...     return {v: k for k, v in d.items()}
        >>> pc.Dict({'a': 1, 'b': 2}).apply(invert_dict)
        {1: 'a', 2: 'b'}

        ```
        """

        def _(data: dict[K, V]) -> dict[KU, VU]:
            return func(data, *args, **kwargs)

        return self._new(_)



from __future__ import annotations

from collections.abc import Callable, Mapping
from functools import partial
from typing import TYPE_CHECKING, Any, TypeIs, overload

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from ._main import Dict


class FilterDict[K, V](MappingWrapper[K, V]):
    @overload
    def filter_keys[U](self, predicate: Callable[[K], TypeIs[U]]) -> Dict[U, V]: ...
    @overload
    def filter_keys(self, predicate: Callable[[K], bool]) -> Dict[K, V]: ...
    def filter_keys[U](
        self,
        predicate: Callable[[K], bool | TypeIs[U]],
    ) -> Dict[K, V] | Dict[U, V]:
        """Return keys that satisfy predicate.

        Args:
            predicate (Callable[[K], bool | TypeIs[U]]): Function to determine if a key should be included.

        Returns:
            Dict[K, V] | Dict[U, V]: Filtered Dict with keys satisfying predicate.

        Example:
        ```python
        >>> import pyochain as pc
        >>> d = {1: 2, 2: 3, 3: 4, 4: 5}
        >>> pc.Dict(d).filter_keys(lambda x: x % 2 == 0).inner()
        {2: 3, 4: 5}

        ```
        """
        return self._new(partial(cz.dicttoolz.keyfilter, predicate))

    @overload
    def filter_values[U](self, predicate: Callable[[V], TypeIs[U]]) -> Dict[K, U]: ...
    @overload
    def filter_values(self, predicate: Callable[[V], bool]) -> Dict[K, V]: ...
    def filter_values[U](
        self,
        predicate: Callable[[V], bool] | Callable[[V], TypeIs[U]],
    ) -> Dict[K, V] | Dict[K, U]:
        """Return items whose values satisfy predicate.

        Args:
            predicate (Callable[[V], bool] | Callable[[V], TypeIs[U]]): Function to determine if a value should be included.

        Returns:
            Dict[K, V] | Dict[K, U]: Filtered Dict with values satisfying predicate

        Example:
        ```python
        >>> import pyochain as pc
        >>> d = {1: 2, 2: 3, 3: 4, 4: 5}
        >>> pc.Dict(d).filter_values(lambda x: x % 2 == 0).inner()
        {1: 2, 3: 4}
        >>> pc.Dict(d).filter_values(lambda x: not x > 3).inner()
        {1: 2, 2: 3}

        ```
        """
        return self._new(partial(cz.dicttoolz.valfilter, predicate))

    def filter_items(self, predicate: Callable[[tuple[K, V]], bool]) -> Dict[K, V]:
        """Filter items by predicate applied to (key, value) tuples.

        Args:
            predicate (Callable[[tuple[K, V]], bool]): Function to determine if a (key, value) pair should be included.

        Returns:
            Dict[K, V]: A new Dict instance containing only the items that satisfy the predicate.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def isvalid(item):
        ...     k, v = item
        ...     return k % 2 == 0 and v < 4
        >>> d = pc.Dict({1: 2, 2: 3, 3: 4, 4: 5})
        >>>
        >>> d.filter_items(isvalid).inner()
        {2: 3}
        >>> d.filter_items(lambda kv: not isvalid(kv)).inner()
        {1: 2, 3: 4, 4: 5}

        ```
        """
        return self._new(partial(cz.dicttoolz.itemfilter, predicate))

    def filter_kv(self, predicate: Callable[[K, V], bool]) -> Dict[K, V]:
        """Filter items by predicate applied to unpacked (key, value) tuples.

        Args:
            predicate(Callable[[K, V], bool]): Function to determine if a key-value pair should be included.

        Returns:
            Dict[K, V]: Filtered Dict with items satisfying predicate.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def isvalid(key, value):
        ...     return key % 2 == 0 and value < 4
        >>> d = pc.Dict({1: 2, 2: 3, 3: 4, 4: 5})
        >>>
        >>> d.filter_kv(isvalid).inner()
        {2: 3}
        >>> d.filter_kv(lambda k, v: not isvalid(k, v)).inner()
        {1: 2, 3: 4, 4: 5}

        ```
        """

        def _filter_kv(data: dict[K, V]) -> dict[K, V]:
            def _(kv: tuple[K, V]) -> bool:
                return predicate(kv[0], kv[1])

            return cz.dicttoolz.itemfilter(_, data)

        return self._new(_filter_kv)

    def filter_attr[U](self, attr: str, dtype: type[U] = object) -> Dict[K, U]:  # noqa: ARG002
        """Filter values that have a given attribute.

        This does not enforce type checking at runtime for performance considerations.

        Args:
            attr (str): Attribute name to check for.
            dtype (type[U]): Optional expected type of the attribute for type hinting.

        Returns:
            Dict[K, U]: A new Dict instance containing only the items whose values have the specified attribute.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"a": "hello", "b": "world", "c": 2, "d": 5}).filter_attr(
        ...     "capitalize", str
        ... ).inner()
        {'a': 'hello', 'b': 'world'}

        ```
        """

        def _filter_attr(data: dict[K, V]) -> dict[K, U]:
            def has_attr(x: object) -> TypeIs[U]:
                return hasattr(x, attr)

            return cz.dicttoolz.valfilter(has_attr, data)

        return self._new(_filter_attr)

    def filter_type[R](self, dtype: type[R]) -> Dict[K, R]:
        """Filter values by type.

        Args:
            dtype (type[R]): Type to filter values by.

        Returns:
            Dict[K, R]: A new Dict instance containing only the items whose values are of the

        Example:
        ```python
        >>> import pyochain as pc
        >>> data = {"a": "one", "b": "two", "c": 3, "d": 4}
        >>> pc.Dict(data).filter_type(str).inner()
        {'a': 'one', 'b': 'two'}

        ```
        """

        def _filter_type(data: dict[K, V]) -> dict[K, R]:
            def _(x: object) -> TypeIs[R]:
                return isinstance(x, dtype)

            return cz.dicttoolz.valfilter(_, data)

        return self._new(_filter_type)

    def filter_subclass[U: type[Any], R](
        self: FilterDict[K, U],
        parent: type[R],
        *,
        keep_parent: bool = True,
    ) -> Dict[K, type[R]]:
        """Filter values that are subclasses of a given parent class.

        Args:
            parent (type[R]): Parent class to check against.
            keep_parent (bool): Whether to include the parent class itself. Defaults to True.

        Returns:
            Dict[K, type[R]]: A new Dict instance containing only the items whose values are subclasses of the specified parent class.

        ```python
        >>> import pyochain as pc
        >>> class A:
        ...     pass
        >>> class B(A):
        ...     pass
        >>> class C:
        ...     pass
        >>> def name(cls: type[Any]) -> str:
        ...     return cls.__name__
        >>> data = pc.Dict({"first": A, "second": B, "third": C})
        >>> data.filter_subclass(A).map_values(name).inner()
        {'first': 'A', 'second': 'B'}
        >>> data.filter_subclass(A, keep_parent=False).map_values(name).inner()
        {'second': 'B'}

        ```
        """

        def _filter_subclass(data: dict[K, U]) -> dict[K, type[R]]:
            def _(x: type[Any]) -> TypeIs[type[R]]:
                if keep_parent:
                    return issubclass(x, parent)
                return issubclass(x, parent) and x is not parent

            return cz.dicttoolz.valfilter(_, data)

        return self._new(_filter_subclass)

    def intersect_keys(self, *others: Mapping[K, V]) -> Dict[K, V]:
        """Keep only keys present in self and all others mappings.

        Args:
            *others (Mapping[K, V]): Other mappings to intersect keys with.

        Returns:
            Dict[K, V]: A new Dict with only the intersected keys.

        ```python
        >>> import pyochain as pc
        >>> d1 = {"a": 1, "b": 2, "c": 3}
        >>> d2 = {"b": 10, "c": 20}
        >>> d3 = {"c": 30}
        >>> pc.Dict(d1).intersect_keys(d2, d3).inner()
        {'c': 3}

        ```
        """

        def _intersect_keys(data: dict[K, V]) -> dict[K, V]:
            self_keys = set(data.keys())
            for other in others:
                self_keys.intersection_update(other.keys())
            return {k: data[k] for k in self_keys}

        return self._new(_intersect_keys)

    def diff_keys(self, *others: Mapping[K, V]) -> Dict[K, V]:
        """Keep only keys present in self but not in others mappings.

        Args:
            *others (Mapping[K, V]): Other mappings to exclude keys from.

        Returns:
            Dict[K, V]: A new Dict with only the differing keys.
        ```python
        >>> import pyochain as pc
        >>> d1 = {"a": 1, "b": 2, "c": 3}
        >>> d2 = {"b": 10, "d": 40}
        >>> d3 = {"c": 30}
        >>> pc.Dict(d1).diff_keys(d2, d3).inner()
        {'a': 1}

        ```
        """

        def _diff_keys(data: dict[K, V]) -> dict[K, V]:
            self_keys = set(data.keys())
            for other in others:
                self_keys.difference_update(other.keys())
            return {k: data[k] for k in self_keys}

        return self._new(_diff_keys)



from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from ._main import Dict


class MapDict[K, V](MappingWrapper[K, V]):
    def map_keys[T](self, func: Callable[[K], T]) -> Dict[T, V]:
        """Return keys transformed by func.

        Args:
            func (Callable[[K], T]): Function to apply to each key in the dictionary.

        Returns:
            Dict[T, V]: Dict with transformed keys.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"Alice": [20, 15, 30], "Bob": [10, 35]}).map_keys(
        ...     str.lower
        ... ).inner()
        {'alice': [20, 15, 30], 'bob': [10, 35]}
        >>>
        >>> pc.Dict({1: "a"}).map_keys(str).inner()
        {'1': 'a'}

        ```
        """
        return self._new(partial(cz.dicttoolz.keymap, func))

    def map_values[T](self, func: Callable[[V], T]) -> Dict[K, T]:
        """Return values transformed by func.

        Args:
            func (Callable[[V], T]): Function to apply to each value in the dictionary.

        Returns:
            Dict[K, T]: Dict with transformed values.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"Alice": [20, 15, 30], "Bob": [10, 35]}).map_values(sum).inner()
        {'Alice': 65, 'Bob': 45}
        >>>
        >>> pc.Dict({1: 1}).map_values(lambda v: v + 1).inner()
        {1: 2}

        ```
        """
        return self._new(partial(cz.dicttoolz.valmap, func))

    def map_items[KR, VR](
        self,
        func: Callable[[tuple[K, V]], tuple[KR, VR]],
    ) -> Dict[KR, VR]:
        """Transform (key, value) pairs using a function that takes a (key, value) tuple.

        Args:
            func (Callable[[tuple[K, V]], tuple[KR, VR]]): Function to transform each (key, value) pair into a new (key, value) tuple.

        Returns:
            Dict[KR, VR]: Dict with transformed items.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({"Alice": 10, "Bob": 20}).map_items(
        ...     lambda kv: (kv[0].upper(), kv[1] * 2)
        ... ).inner()
        {'ALICE': 20, 'BOB': 40}

        ```
        """
        return self._new(partial(cz.dicttoolz.itemmap, func))

    def map_kv[KR, VR](
        self,
        func: Callable[[K, V], tuple[KR, VR]],
    ) -> Dict[KR, VR]:
        """Transform (key, value) pairs using a function that takes key and value as separate arguments.

        Args:
            func (Callable[[K, V], tuple[KR, VR]]): Function to transform each key and value into a new (key, value) tuple.

        Returns:
            Dict[KR, VR]: Dict with transformed items.

        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2}).map_kv(lambda k, v: (k + 1, v * 10)).inner()
        {2: 20}

        ```
        """

        def _map_kv(data: dict[K, V]) -> dict[KR, VR]:
            def _(kv: tuple[K, V]) -> tuple[KR, VR]:
                return func(kv[0], kv[1])

            return cz.dicttoolz.itemmap(_, data)

        return self._new(_map_kv)

    def invert(self) -> Dict[V, list[K]]:
        """Invert the dictionary, grouping keys by common (and hashable) values.

        Returns:
            Dict[V, list[K]]: Inverted Dict with values as keys and lists of original keys as values.

        ```python
        >>> import pyochain as pc
        >>> d = {"a": 1, "b": 2, "c": 1}
        >>> pc.Dict(d).invert().inner()
        {1: ['a', 'c'], 2: ['b']}

        ```
        """

        def _invert(data: dict[K, V]) -> dict[V, list[K]]:
            inverted: dict[V, list[K]] = defaultdict(list)
            for k, v in data.items():
                inverted[v].append(k)
            return dict(inverted)

        return self._new(_invert)

    def implode(self) -> Dict[K, list[V]]:
        """Nest all the values in lists.

        Returns:
            Dict[K, list[V]]: Dict with all values nested in lists.

        syntactic sugar for map_values(lambda v: [v])
        ```python
        >>> import pyochain as pc
        >>> pc.Dict({1: 2, 3: 4}).implode().inner()
        {1: [2], 3: [4]}

        ```
        """

        def _implode(data: dict[K, V]) -> dict[K, list[V]]:
            def _(v: V) -> list[V]:
                return [v]

            return cz.dicttoolz.valmap(_, data)

        return self._new(_implode)



from __future__ import annotations

import itertools
from collections.abc import (
    Callable,
    Generator,
    Iterable,
    Iterator,
    Sequence,
)
from typing import TYPE_CHECKING, Any, Concatenate, TypeIs, overload, override
from warnings import deprecated

import cytoolz as cz

from ._aggregations import BaseAgg
from ._booleans import BaseBool
from ._dicts import BaseDict
from ._eager import BaseEager
from ._filters import BaseFilter
from ._joins import BaseJoins
from ._lists import BaseList
from ._maps import BaseMap
from ._partitions import BasePartitions
from ._process import BaseProcess
from ._rolling import BaseRolling
from ._tuples import BaseTuples

if TYPE_CHECKING:
    from .._dict import Dict


class CommonMethods[T](BaseAgg[T], BaseEager[T], BaseDict[T]):
    pass


def _convert_data[T](data: Iterable[T] | T, *more_data: T) -> Iterable[T]:
    return data if cz.itertoolz.isiterable(data) else (data, *more_data)


def _is_sequence[T](data: Iterable[T]) -> TypeIs[Sequence[T]]:
    return hasattr(data, "__getitem__") and hasattr(data, "__len__")


class Iter[T](
    BaseBool[T],
    BaseFilter[T],
    BaseProcess[T],
    BaseMap[T],
    BaseRolling[T],
    BaseList[T],
    BaseTuples[T],
    BasePartitions[T],
    BaseJoins[T],
    CommonMethods[T],
):
    """A wrapper around Python's built-in `Iterators`/`Generators` types, providing a rich set of functional programming tools.

    - An `Iterable` is any object capable of returning its members one at a time, permitting it to be iterated over in a for-loop.
    - An `Iterator` is an object representing a stream of data; returned by calling `iter()` on an `Iterable`.
    - Once an `Iterator` is exhausted, it cannot be reused or reset.

    It's designed around lazy evaluation, allowing for efficient processing of large datasets.

    - To instantiate from a lazy `Iterator`/`Generator`, simply pass it to the standard constructor.
    - To instantiate from any `Iterable` (like a list or set), or unpacked values, use the `from_` class method.

    Once an `Iter` is created, it can be transformed and manipulated using a variety of chainable methods.

    However, keep in mind that `Iter` instances are single-use; once exhausted, they cannot be reused or reset.

    If you need to reuse the data, consider collecting it into a `Seq` first with `.collect()`.

    You can always convert back to an `Iter` using `Seq.iter()` for free.

    In general, avoid intermediate references when dealing with lazy iterators, and prioritize method chaining instead.

    Args:
        data (Iterator[T] | Generator[T, Any, Any]): An iterator or generator to wrap.
    """

    __slots__ = ("_inner",)

    def __init__(self, data: Iterator[T] | Generator[T, Any, Any]) -> None:
        self._inner = data

    def next(self) -> T:
        """Call next builtin on the underlying iterator to get the next element.

        Returns:
            T: The next element in the iterator.

        Example:
        ```python
        >>> import pyochain as pc
        >>> it = pc.Iter.from_([1, 2, 3])
        >>> it.next()
        1
        >>> it.next()
        2

        ```
        """
        return next(self.inner())

    @staticmethod
    def from_count(start: int = 0, step: int = 1) -> Iter[int]:
        """Create an infinite `Iterator` of evenly spaced values.

        **Warning** ⚠️
            This creates an infinite iterator.
            Be sure to use `Iter.take()` or `Iter.slice()` to limit the number of items taken.

        Args:
            start (int): Starting value of the sequence. Defaults to 0.
            step (int): Difference between consecutive values. Defaults to 1.

        Returns:
            Iter[int]: An iterator generating the sequence.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_count(10, 2).take(3).into(list)
        [10, 12, 14]

        ```
        """
        return Iter(itertools.count(start, step))

    @staticmethod
    def from_func[U](func: Callable[[U], U], value: U) -> Iter[U]:
        """Create an infinite iterator by repeatedly applying a function on an original value.

        **Warning** ⚠️
            This creates an infinite iterator.
            Be sure to use `Iter.take()` or `Iter.slice()` to limit the number of items taken.

        Args:
            func (Callable[[U], U]): Function to apply repeatedly.
            value (U): Initial value to start the iteration.

        Returns:
            Iter[U]: An iterator generating the sequence.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_func(lambda x: x + 1, 0).take(3).into(list)
        [0, 1, 2]

        ```
        """
        return Iter(cz.itertoolz.iterate(func, value))

    @overload
    @staticmethod
    def from_[U](data: Iterable[U]) -> Iter[U]: ...
    @overload
    @staticmethod
    def from_[U](data: U, *more_data: U) -> Iter[U]: ...
    @staticmethod
    def from_[U](data: Iterable[U] | U, *more_data: U) -> Iter[U]:
        """Create an iterator from any Iterable, or from unpacked values.

        Prefer using the standard constructor, as this method involves extra checks and conversions steps.

        Args:
            data (Iterable[U] | U): Iterable to convert into an iterator, or a single value.
            *more_data (U): Additional values to include if 'data' is not an Iterable.

        Returns:
            Iter[U]: A new Iter instance containing the provided data.

        Example:
        ```python
        >>> import pyochain as pc
        >>> data: tuple[int, ...] = (1, 2, 3)
        >>> iterator = pc.Iter.from_(data)
        >>> iterator.inner().__class__.__name__
        'tuple_iterator'
        >>> mapped = iterator.map(lambda x: x * 2)
        >>> mapped.inner().__class__.__name__
        'map'
        >>> mapped.collect(tuple)
        Seq((2, 4, 6))
        >>> # iterator is now exhausted
        >>> iterator.collect()
        Seq([])
        >>> # Creating from unpacked values
        >>> pc.Iter.from_(1, 2, 3).collect(tuple)
        Seq((1, 2, 3))

        ```
        """
        return Iter(iter(_convert_data(data, *more_data)))

    @staticmethod
    def unfold[S, V](seed: S, generator: Callable[[S], tuple[V, S] | None]) -> Iter[V]:
        """Create an iterator by repeatedly applying a generator function to an initial state.

        The `generator` function takes the current state and must return:

        - A tuple `(value, new_state)` to emit the `value` and continue with the `new_state`.
        - `None` to stop the generation.

        This is functionally equivalent to a state-based `while` loop.

        **Warning** ⚠️
            If the `generator` function never returns `None`, it creates an infinite iterator.
            Be sure to use `Iter.take()` or `Iter.slice()` to limit the number of items taken if necessary.

        Args:
            seed (S): Initial state for the generator.
            generator (Callable[[S], tuple[V, S] | None]): Function that generates the next value and state.

        Returns:
            Iter[V]: An iterator generating values produced by the generator function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> # Example 1: Simple counter up to 5
        >>> def counter_generator(state: int) -> tuple[int, int] | None:
        ...     if state < 5:
        ...         return (state * 10, state + 1)
        ...     return None
        >>> pc.Iter.unfold(seed=0, generator=counter_generator).into(list)
        [0, 10, 20, 30, 40]
        >>> # Example 2: Fibonacci sequence up to 100
        >>> type FibState = tuple[int, int]
        >>> def fib_generator(state: FibState) -> tuple[int, FibState] | None:
        ...     a, b = state
        ...     if a > 100:
        ...         return None
        ...     return (a, (b, a + b))
        >>> pc.Iter.unfold(seed=(0, 1), generator=fib_generator).into(list)
        [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89]
        >>> # Example 3: Infinite iterator (requires take())
        >>> pc.Iter.unfold(seed=1, generator=lambda s: (s, s * 2)).take(5).into(list)
        [1, 2, 4, 8, 16]

        ```
        """
        from ._main import Iter

        def _unfold() -> Iterator[V]:
            current_seed: S = seed
            while True:
                result: tuple[V, S] | None = generator(current_seed)
                if result is None:
                    break
                value, next_seed = result
                yield value
                current_seed = next_seed

        return Iter(_unfold())

    def itr[**P, R, U: Iterable[Any]](
        self: Iter[U],
        func: Callable[Concatenate[Iter[U], P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iter[R]:
        """Apply a function to each element after wrapping it in an Iter.

        This is a convenience method for the common pattern of mapping a function over an iterable of iterables.

        Args:
            func (Callable[Concatenate[Iter[U], P], R]): Function to apply to each wrapped element.
            *args (P.args): Positional arguments to pass to the function.
            **kwargs (P.kwargs): Keyword arguments to pass to the function.

        Returns:
            Iter[R]: A new `Iter` instance containing the results of applying the function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> data = [
        ...     [1, 2, 3],
        ...     [4, 5],
        ...     [6, 7, 8, 9],
        ... ]
        >>> pc.Iter.from_(data).itr(
        ...     lambda x: x.repeat(2).flatten().reduce(lambda a, b: a + b)
        ... ).into(list)
        [12, 18, 60]

        ```
        """

        def _itr(data: Iterable[U]) -> Generator[R, None, None]:
            return (func(Iter(iter(x)), *args, **kwargs) for x in data)

        return self._lazy(_itr)

    def struct[**P, R, K, V](
        self: Iter[dict[K, V]],
        func: Callable[Concatenate[Dict[K, V], P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iter[R]:
        """Apply a function to each element after wrapping it in a `Dict`.

        This is a convenience method for the common pattern of mapping a function over an `Iterable` of dictionaries.

        Args:
            func (Callable[Concatenate[Dict[K, V], P], R]): Function to apply to each wrapped dictionary.
            *args (P.args): Positional arguments to pass to the function.
            **kwargs (P.kwargs): Keyword arguments to pass to the function.

        Returns:
            Iter[R]: A new `Iter` instance containing the results of applying the function.

        Example:
        ```python
        >>> from typing import Any
        >>> import pyochain as pc

        >>> data: list[dict[str, Any]] = [
        ...     {"name": "Alice", "age": 30, "city": "New York"},
        ...     {"name": "Bob", "age": 25, "city": "Los Angeles"},
        ...     {"name": "Charlie", "age": 35, "city": "New York"},
        ...     {"name": "David", "age": 40, "city": "Paris"},
        ... ]
        >>>
        >>> def to_title(d: pc.Dict[str, Any]) -> pc.Dict[str, Any]:
        ...     return d.map_keys(lambda k: k.title())
        >>>
        >>> def is_young(d: pc.Dict[str, Any]) -> bool:
        ...     return d.inner().get("Age", 0) < 30
        >>>
        >>> def set_continent(d: pc.Dict[str, Any], value: str) -> dict[str, Any]:
        ...     return d.with_key("Continent", value).inner()
        >>>
        >>> def grouped_data():
        ...     return (
        ...         pc.Iter.from_(data)
        ...         .struct(to_title)
        ...         .filter_false(is_young)
        ...         .map(lambda d: d.drop("Age").with_key("Continent", "NA"))
        ...         .map_if(
        ...             lambda d: d.inner().get("City") == "Paris",
        ...             lambda d: set_continent(d, "Europe"),
        ...             lambda d: set_continent(d, "America"),
        ...         )
        ...         .group_by(lambda d: d.get("Continent"))
        ...         .map_values(
        ...             lambda d: pc.Iter.from_(d)
        ...             .struct(lambda d: d.drop("Continent").inner())
        ...             .into(list)
        ...         )
        ...     )
        >>> grouped_data()  # doctest: +NORMALIZE_WHITESPACE
        {'America': [{'City': 'New York', 'Name': 'Alice'},
                    {'City': 'New York', 'Name': 'Charlie'}],
        'Europe': [{'City': 'Paris', 'Name': 'David'}]}

        ```
        """
        from .._dict import Dict

        def _struct(data: Iterable[dict[K, V]]) -> Generator[R, None, None]:
            return (func(Dict(x), *args, **kwargs) for x in data)

        return self._lazy(_struct)

    def apply[**P, R](
        self,
        func: Callable[Concatenate[Iterable[T], P], Iterator[R]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Iter[R]:
        """Apply a function to the underlying Iterator and return a new Iter instance.

        Allow to pass user defined functions that transform the iterable while retaining the Iter wrapper.

        Args:
            func (Callable[Concatenate[Iterable[T], P], Iterator[R]]): Function to apply to the underlying iterable.
            *args (P.args): Positional arguments to pass to the function.
            **kwargs (P.kwargs): Keyword arguments to pass to the function.

        Returns:
            Iter[R]: A new `Iter` instance containing the result of the function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def double(data: Iterable[int]) -> Iterator[int]:
        ...     return (x * 2 for x in data)
        >>> pc.Iter.from_([1, 2, 3]).apply(double).into(list)
        [2, 4, 6]

        ```
        """
        return self._lazy(func, *args, **kwargs)

    def collect(self, factory: Callable[[Iterable[T]], Sequence[T]] = list) -> Seq[T]:
        """Collect the elements into a `Sequence`, using the provided factory.

        Args:
            factory (Callable[[Iterable[T]], Sequence[T]]): A callable that takes an iterable and returns a Sequence. Defaults to `list`.

        Returns:
            Seq[T]: A `Seq` containing the collected elements.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(range(5)).collect()
        Seq([0, 1, 2, 3, 4])

        ```
        """
        return self._eager(factory)

    @override
    @deprecated("Use .inner() instead")
    def unwrap(self) -> Iterator[T]:
        return self._inner  # type: ignore[return-value]

    @override
    def inner(self) -> Iterator[T]:
        return self._inner  # type: ignore[return-value]


class Seq[T](CommonMethods[T]):
    """`Seq` represent an in memory Sequence.

    Provides a subset of `Iter` methods with eager evaluation, and is the return type of `Iter.collect()`.

    You can create a `Seq` from any `Sequence` (like a list or a tuple) using the standard constructor,
    or from unpacked values using the `from_` class method.

    Doing `Seq(...).iter()` or `Iter.from_(...)` are equivalent.

    Args:
            data (Sequence[T]): The data to initialize the Seq with.
    """

    __slots__ = ("_inner",)

    def __init__(self, data: Sequence[T]) -> None:
        self._inner = data

    @overload
    @staticmethod
    def from_[U](data: Iterable[U]) -> Seq[U]: ...
    @overload
    @staticmethod
    def from_[U](data: U, *more_data: U) -> Seq[U]: ...
    @staticmethod
    def from_[U](data: Iterable[U] | U, *more_data: U) -> Seq[U]:
        """Create a `Seq` from an `Iterable` or unpacked values.

        Prefer using the standard constructor, as this method involves extra checks and conversions steps.

        Args:
            data (Iterable[U] | U): Iterable to convert into a sequence, or a single value.
            *more_data (U): Unpacked items to include in the sequence, if 'data' is not an Iterable.

        Returns:
            Seq[U]: A new Seq instance containing the provided data.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Seq.from_(1, 2, 3)
        Seq((1, 2, 3))

        ```

        """
        converted = _convert_data(data, *more_data)
        return Seq(converted if _is_sequence(converted) else tuple(converted))

    def iter(self) -> Iter[T]:
        """Get an iterator over the sequence.

        Call this to switch to lazy evaluation.

        Returns:
            Iter[T]: An `Iter` instance wrapping an iterator over the sequence.
        """
        return self._lazy(iter)

    def apply[**P, R](
        self,
        func: Callable[Concatenate[Iterable[T], P], Sequence[R]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Seq[R]:
        """Apply a function to the underlying `Sequence` and return a `Seq` instance.

        Allow to pass user defined functions that transform the `Sequence` while retaining the `Seq` wrapper.

        Args:
            func (Callable[Concatenate[Iterable[T], P], Sequence[R]]): Function to apply to the underlying `Sequence`.
            *args (P.args): Positional arguments to pass to the function.
            **kwargs (P.kwargs): Keyword arguments to pass to the function.

        Returns:
            Seq[R]: A new `Seq` instance containing the result of the function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def double(data: Iterable[int]) -> Sequence[int]:
        ...     return [x * 2 for x in data]
        >>> pc.Seq([1, 2, 3]).apply(double).into(list)
        [2, 4, 6]

        ```
        """
        return self._eager(func, *args, **kwargs)

    @override
    def inner(self) -> Sequence[T]:
        return self._inner  # type: ignore[return-value]

    @override
    @deprecated("Use .inner() instead")
    def unwrap(self) -> Sequence[T]:
        return self._inner  # type: ignore[return-value]

    @override
    def into[**P, R](
        self,
        func: Callable[[Sequence[T]], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        return super().into(func, *args, **kwargs)  # type: ignore[return-value]



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, Any

from ._filters import FilterDict
from ._groups import GroupsDict
from ._iter import IterDict
from ._joins import JoinsDict
from ._maps import MapDict
from ._nested import NestedDict
from ._process import ProcessDict

if TYPE_CHECKING:
    from .._core import SupportsKeysAndGetItem


class DictCommonMethods[K, V](
    ProcessDict[K, V],
    IterDict[K, V],
    NestedDict[K, V],
    MapDict[K, V],
    JoinsDict[K, V],
    FilterDict[K, V],
    GroupsDict[K, V],
):
    pass


class Dict[K, V](DictCommonMethods[K, V]):
    """Wrapper for Python dictionaries with chainable methods."""

    __slots__ = ()

    @staticmethod
    def from_[G, I](
        data: Mapping[G, I] | Iterable[tuple[G, I]] | SupportsKeysAndGetItem[G, I],
    ) -> Dict[G, I]:
        """Create a `Dict` from a convertible value.

        Args:
            data (Mapping[G, I] | Iterable[tuple[G, I]] | SupportsKeysAndGetItem[G, I]): Object convertible into a Dict.

        Returns:
            Dict[G, I]: Instance containing the data from the input.

        Example:
        ```python
        >>> import pyochain as pc
        >>> class MyMapping:
        ...     def __init__(self):
        ...         self._data = {1: "a", 2: "b", 3: "c"}
        ...
        ...     def keys(self):
        ...         return self._data.keys()
        ...
        ...     def __getitem__(self, key):
        ...         return self._data[key]
        >>>
        >>> pc.Dict.from_(MyMapping()).inner()
        {1: 'a', 2: 'b', 3: 'c'}
        >>> pc.Dict.from_([("d", "e"), ("f", "g")]).inner()
        {'d': 'e', 'f': 'g'}

        ```
        """
        return Dict(dict(data))

    @staticmethod
    def from_object(obj: object) -> Dict[str, Any]:
        """Create a `Dict` from an object `__dict__` attribute.

        We can't know in advance the values types, so we use `Any`.

        Args:
            obj (object): The object whose `__dict__` attribute will be used to create the Dict.

        Returns:
            Dict[str, Any]: A new Dict instance containing the attributes of the object.

        ```python
        >>> import pyochain as pc
        >>> class Person:
        ...     def __init__(self, name: str, age: int):
        ...         self.name = name
        ...         self.age = age
        >>> person = Person("Alice", 30)
        >>> pc.Dict.from_object(person).inner()
        {'name': 'Alice', 'age': 30}

        ```
        """
        return Dict(obj.__dict__)

    def pivot(self, *indices: int) -> Dict[Any, Any]:
        """Pivot a nested dictionary by rearranging the key levels according to order.

        Syntactic sugar for `Dict.to_arrays().rearrange(*indices).to_records()`

        Args:
            *indices (int): Indices specifying the new order of key levels

        Returns:
            Dict[Any, Any]: Pivoted dictionary with keys rearranged

        Example:
        ```python
        >>> import pyochain as pc
        >>> d = {"A": {"X": 1, "Y": 2}, "B": {"X": 3, "Y": 4}}
        >>> pc.Dict(d).pivot(1, 0).inner()
        {'X': {'A': 1, 'B': 3}, 'Y': {'A': 2, 'B': 4}}
        """
        return self.to_arrays().rearrange(*indices).to_records()



from ._option import Option
from ._result import Result, ResultUnwrapError
from ._states import NONE, Err, Ok, OptionUnwrapError, Some, Wrapper

__all__ = [
    "NONE",
    "Err",
    "Ok",
    "Option",
    "OptionUnwrapError",
    "Result",
    "ResultUnwrapError",
    "Some",
    "Wrapper",
]



from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from functools import partial
from typing import TYPE_CHECKING, Any

import cytoolz as cz

from .._core import IterWrapper, SupportsRichComparison

if TYPE_CHECKING:
    from ._main import Seq


class BaseEager[T](IterWrapper[T]):
    def sort[U: SupportsRichComparison[Any]](
        self: BaseEager[U],
        key: Callable[[U], Any] | None = None,
        *,
        reverse: bool = False,
    ) -> Seq[U]:
        """Sort the elements of the sequence.

        Note:
            This method must consume the entire iterable to perform the sort.
            The result is a new iterable over the sorted sequence.

        Args:
            key (Callable[[U], Any] | None): Function to extract a comparison key from each element. Defaults to None.
            reverse (bool): Whether to sort in descending order. Defaults to False.

        Returns:
            Seq[U]: A new Seq with elements sorted.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Seq([3, 1, 2]).sort()
        Seq([1, 2, 3])

        ```
        """

        def _sort(data: Iterable[U]) -> list[U]:
            return sorted(data, reverse=reverse, key=key)

        return self._eager(_sort)

    def tail(self, n: int) -> Seq[T]:
        """Return a tuple of the last n elements.

        Args:
            n (int): Number of elements to return.

        Returns:
            Seq[T]: A new Seq containing the last n elements.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).tail(2)
        Seq((2, 3))

        ```
        """
        return self._eager(partial(cz.itertoolz.tail, n))

    def top_n(self, n: int, key: Callable[[T], Any] | None = None) -> Seq[T]:
        """Return a tuple of the top-n items according to key.

        Args:
            n (int): Number of top elements to return.
            key (Callable[[T], Any] | None): Function to extract a comparison key from each element. Defaults to None.

        Returns:
            Seq[T]: A new Seq containing the top-n elements.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 3, 2]).top_n(2)
        Seq((3, 2))

        ```
        """
        return self._eager(partial(cz.itertoolz.topk, n, key=key))

    def union(self, *others: Iterable[T]) -> Seq[T]:
        """Return the union of this iterable and 'others'.

        Note:
            This method consumes inner data and removes duplicates.

        Args:
            *others (Iterable[T]): Other iterables to include in the union.

        Returns:
            Seq[T]: A new Seq containing the union of elements.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 2]).union([2, 3], [4]).iter().sort()
        Seq([1, 2, 3, 4])

        ```
        """

        def _union(data: Iterable[T]) -> list[T]:
            return list(set(data).union(*others))

        return self._eager(_union)

    def intersection(self, *others: Iterable[T]) -> Seq[T]:
        """Return the elements common to this iterable and 'others'.

        Note:
            This method consumes inner data, unsorts it, and removes duplicates.

        Args:
            *others (Iterable[T]): Other iterables to intersect with.

        Returns:
            Seq[T]: A new Seq containing the intersection of elements.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 2]).intersection([2, 3], [2])
        Seq([2])

        ```
        """

        def _intersection(data: Iterable[T]) -> list[T]:
            return list(set(data).intersection(*others))

        return self._eager(_intersection)

    def diff_unique(self, *others: Iterable[T]) -> Seq[T]:
        """Return the difference of this iterable and 'others'.

        (Elements in 'self' but not in 'others').

        Note:
            This method consumes inner data, unsorts it, and removes duplicates.

        Args:
            *others (Iterable[T]): Other iterables to subtract from this iterable.

        Returns:
            Seq[T]: A new Seq containing the difference of elements.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 2]).diff_unique([2, 3])
        Seq([1])

        ```
        """

        def _difference(data: Iterable[T]) -> list[T]:
            return list(set(data).difference(*others))

        return self._eager(_difference)

    def diff_symmetric(self, *others: Iterable[T]) -> Seq[T]:
        """Return the symmetric difference (XOR) of this iterable and 'others'.

        Note:
            This method consumes inner data, unsorts it, and removes duplicates.

        Args:
            *others (Iterable[T]): Other iterables to compute the symmetric difference with.

        Returns:
            Seq[T]: A new Seq containing the symmetric difference of elements.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 2]).diff_symmetric([2, 3]).iter().sort()
        Seq([1, 3])
        >>> pc.Iter.from_([1, 2, 3]).diff_symmetric([3, 4, 5]).iter().sort()
        Seq([1, 2, 4, 5])

        ```
        """

        def _symmetric_difference(data: Iterable[T]) -> list[T]:
            return list(set(data).symmetric_difference(*others))

        return self._eager(_symmetric_difference)

    def most_common(self, n: int | None = None) -> Seq[tuple[T, int]]:
        """Return the n most common elements and their counts.

        If n is None, then all elements are returned.

        Args:
            n (int | None): Number of most common elements to return. Defaults to None (all elements).

        Returns:
            Seq[tuple[T, int]]: A new Seq containing tuples of (element, count).

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 1, 2, 3, 3, 3]).most_common(2)
        Seq([(3, 3), (1, 2)])

        ```
        """
        from collections import Counter

        def _most_common(data: Iterable[T]) -> list[tuple[T, int]]:
            return Counter(data).most_common(n)

        return self._eager(_most_common)

    def rearrange[U: Sequence[Any]](self: BaseEager[U], *indices: int) -> Seq[list[U]]:
        """Rearrange elements in a given list of arrays by order indices.

        The last element (value) always remains in place.

        Args:
            *indices (int): indices specifying new order of keys

        Raises:
            IndexError: If any index in order is out of range for the row.

        Returns:
            Seq[list[U]]: A new Seq containing rearranged elements.

        Example:
        ```python
        >>> import pyochain as pc
        >>> data = pc.Seq([["A", "X", 1], ["A", "Y", 2], ["B", "X", 3], ["B", "Y", 4]])
        >>> data.rearrange(1, 0)
        Seq([['X', 'A', 1], ['Y', 'A', 2], ['X', 'B', 3], ['Y', 'B', 4]])

        ```
        """

        def _check_bound(i: int, max_key_index: int) -> None:
            if i < 0 or i > max_key_index:
                msg = f"order index {i} out of range for row with {max_key_index + 1} keys"
                raise IndexError(
                    msg,
                )

        def _rearrange(in_arrs: Iterable[U]) -> list[list[U]]:
            order = indices
            out: list[list[U]] = []
            for arr in in_arrs:
                max_key_index: int = len(arr) - 2
                for i in order:
                    _check_bound(i, max_key_index)

                out.append([arr[i] for i in order] + [arr[-1]])

            return out

        return self._eager(_rearrange)



from __future__ import annotations

from collections.abc import Callable, Mapping
from functools import partial
from typing import TYPE_CHECKING, Any, Concatenate, TypeIs

import cytoolz as cz

from .._core import MappingWrapper

if TYPE_CHECKING:
    from ._main import Dict


def _drop_nones(
    data: dict[Any, Any] | list[Any],
    *,
    remove_empty: bool = True,
) -> dict[Any, Any] | list[Any] | None:
    match data:
        case dict():
            pruned_dict: dict[Any, Any] = {}
            for k, v in data.items():
                pruned_v = _drop_nones(v, remove_empty=remove_empty)

                is_empty = remove_empty and (pruned_v is None or pruned_v in ({}, []))
                if not is_empty:
                    pruned_dict[k] = pruned_v
            return pruned_dict if pruned_dict or not remove_empty else None

        case list():
            pruned_list = [
                _drop_nones(item, remove_empty=remove_empty) for item in data
            ]
            if remove_empty:
                pruned_list = [
                    item
                    for item in pruned_list
                    if not (item is None or item in ({}, []))
                ]
            return pruned_list if pruned_list or not remove_empty else None

        case _:
            if remove_empty and data is None:
                return None
            return data


class NestedDict[K, V](MappingWrapper[K, V]):
    def struct[**P, R, U: dict[Any, Any]](
        self: NestedDict[K, U],
        func: Callable[Concatenate[Dict[K, U], P], R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Dict[K, R]:
        """Apply a function to each value after wrapping it in a Dict.

        Args:
            func (Callable[Concatenate[Dict[K, U], P], R]): Function to apply to each value after wrapping it in a Dict.
            *args (P.args): Positional arguments to pass to the function.
            **kwargs (P.kwargs): Keyword arguments to pass to the function.

        Returns:
            Dict[K, R]: Dict with function results as values.

        Syntactic sugar for `map_values(lambda data: func(pc.Dict(data), *args, **kwargs))`
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "person1": {"name": "Alice", "age": 30, "city": "New York"},
        ...     "person2": {"name": "Bob", "age": 25, "city": "Los Angeles"},
        ... }
        >>> pc.Dict(data).struct(lambda d: d.map_keys(str.upper).drop("AGE").inner())
        ... # doctest: +NORMALIZE_WHITESPACE
        {'person1': {'CITY': 'New York', 'NAME': 'Alice'},
        'person2': {'CITY': 'Los Angeles', 'NAME': 'Bob'}}

        ```
        """
        from ._main import Dict

        def _struct(data: Mapping[K, U]) -> dict[K, R]:
            def _(v: dict[Any, Any]) -> R:
                return func(Dict(v), *args, **kwargs)

            return cz.dicttoolz.valmap(_, data)

        return self._new(_struct)

    def flatten(
        self: NestedDict[str, Any],
        sep: str = ".",
        max_depth: int | None = None,
    ) -> Dict[str, Any]:
        """Flatten a nested dictionary, concatenating keys with the specified separator.

        Args:
            sep (str): Separator to use when concatenating keys
            max_depth (int | None): Maximum depth to flatten. If None, flattens completely.

        Returns:
            Dict[str, Any]: Flattened Dict with concatenated keys.
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "config": {"params": {"retries": 3, "timeout": 30}, "mode": "fast"},
        ...     "version": 1.0,
        ... }
        >>> pc.Dict(data).flatten().inner()
        {'config.params.retries': 3, 'config.params.timeout': 30, 'config.mode': 'fast', 'version': 1.0}
        >>> pc.Dict(data).flatten(sep="_").inner()
        {'config_params_retries': 3, 'config_params_timeout': 30, 'config_mode': 'fast', 'version': 1.0}
        >>> pc.Dict(data).flatten(max_depth=1).inner()
        {'config.params': {'retries': 3, 'timeout': 30}, 'config.mode': 'fast', 'version': 1.0}

        ```
        """

        def _flatten(
            d: Mapping[Any, Any],
            parent_key: str = "",
            current_depth: int = 1,
        ) -> dict[str, Any]:
            def _can_recurse(v: object) -> TypeIs[Mapping[Any, Any]]:
                return isinstance(v, Mapping) and (
                    max_depth is None or current_depth < max_depth + 1
                )

            items: list[tuple[str, Any]] = []
            for k, v in d.items():
                new_key = parent_key + sep + k if parent_key else k
                if _can_recurse(v):
                    items.extend(_flatten(v, new_key, current_depth + 1).items())
                else:
                    items.append((new_key, v))
            return dict(items)

        return self._new(_flatten)

    def unpivot(
        self: NestedDict[str, Mapping[str, Any]],
    ) -> Dict[str, dict[str, Any]]:
        """Unpivot a nested dictionary by swapping rows and columns.

        Returns:
            Dict[str, dict[str, Any]]: Unpivoted Dict with columns as keys and rows as sub-dicts.

        Example:
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "row1": {"col1": "A", "col2": "B"},
        ...     "row2": {"col1": "C", "col2": "D"},
        ... }
        >>> pc.Dict(data).unpivot()
        ... # doctest: +NORMALIZE_WHITESPACE
        {'col1': {'row1': 'A', 'row2': 'C'}, 'col2': {'row1': 'B', 'row2': 'D'}}
        """

        def _unpivot(
            data: Mapping[str, Mapping[str, Any]],
        ) -> dict[str, dict[str, Any]]:
            out: dict[str, dict[str, Any]] = {}
            for rkey, inner in data.items():
                for ckey, val in inner.items():
                    out.setdefault(ckey, {})[rkey] = val
            return out

        return self._new(_unpivot)

    def with_nested_key(self, *keys: K, value: V) -> Dict[K, V]:
        """Set a nested key path and return a new Dict with new, potentially nested, key value pair.

        Args:
            *keys (K): Sequence of keys representing the nested path.
            value (V): Value to set at the specified nested path.

        Returns:
            Dict[K, V]: Dict with the new nested key-value pair.
        ```python
        >>> import pyochain as pc
        >>> purchase = {
        ...     "name": "Alice",
        ...     "order": {"items": ["Apple", "Orange"], "costs": [0.50, 1.25]},
        ...     "credit card": "5555-1234-1234-1234",
        ... }
        >>> pc.Dict(purchase).with_nested_key(
        ...     "order", "costs", value=[0.25, 1.00]
        ... ).inner()
        {'name': 'Alice', 'order': {'items': ['Apple', 'Orange'], 'costs': [0.25, 1.0]}, 'credit card': '5555-1234-1234-1234'}

        ```
        """

        def _with_nested_key(data: dict[K, V]) -> dict[K, V]:
            return cz.dicttoolz.assoc_in(data, keys, value=value)

        return self._new(_with_nested_key)

    def pluck[U: str | int](self: NestedDict[U, Any], *keys: str) -> Dict[U, Any]:
        """Extract values from nested dictionaries using a sequence of keys.

        Args:
            *keys (str): Sequence of keys to extract values from the nested dictionaries.

        Returns:
            Dict[U, Any]: Dict with extracted values from nested dictionaries.
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "person1": {"name": "Alice", "age": 30},
        ...     "person2": {"name": "Bob", "age": 25},
        ... }
        >>> pc.Dict(data).pluck("name").inner()
        {'person1': 'Alice', 'person2': 'Bob'}

        ```
        """
        getter = partial(cz.dicttoolz.get_in, keys)

        def _pluck(data: Mapping[U, Any]) -> dict[U, Any]:
            return cz.dicttoolz.valmap(getter, data)

        return self._new(_pluck)

    def get_in[T](self, *keys: K, default: T = None) -> V | T:
        """Retrieve a value from a nested dictionary structure.

        Args:
            *keys (K): Sequence of keys representing the nested path to retrieve the value.
            default (T): Default value to return if the keys do not exist.

        Returns:
            V | T: Value at the nested path or default if not found.

        ```python
        >>> import pyochain as pc
        >>> data = {"a": {"b": {"c": 1}}}
        >>> pc.Dict(data).get_in("a", "b", "c")
        1
        >>> pc.Dict(data).get_in("a", "x", default="Not Found")
        'Not Found'

        ```
        """

        def _get_in(data: Mapping[K, V]) -> V | T:
            return cz.dicttoolz.get_in(keys, data, default)

        return self.into(_get_in)

    def drop_nones(self, *, remove_empty: bool = True) -> Dict[K, V]:
        """Recursively drop None values from the dictionary.

        Options to also remove empty dicts and lists.

        Args:
            remove_empty (bool): If True (default), removes `None`, `{}` and `[]`.

        Returns:
            Dict[K, V]: Dict with None values and optionally empty structures removed.

        Example:
        ```python
        >>> import pyochain as pc
        >>> data = {
        ...     "a": 1,
        ...     "b": None,
        ...     "c": {},
        ...     "d": [],
        ...     "e": {"f": None, "g": 2},
        ...     "h": [1, None, {}],
        ...     "i": 0,
        ... }
        >>> p_data = pc.Dict(data)
        >>>
        >>> p_data.drop_nones().inner()
        {'a': 1, 'e': {'g': 2}, 'h': [1], 'i': 0}
        >>>
        >>> p_data.drop_nones().inner()
        {'a': 1, 'e': {'g': 2}, 'h': [1], 'i': 0}
        >>>
        >>> p_data.drop_nones(remove_empty=False).inner()
        {'a': 1, 'b': None, 'c': {}, 'd': [], 'e': {'f': None, 'g': 2}, 'h': [1, None, {}], 'i': 0}

        ```
        """

        def _apply_drop_nones(data: dict[K, V]) -> dict[Any, Any]:
            result = _drop_nones(data, remove_empty=remove_empty)
            return result if isinstance(result, dict) else {}

        return self._new(_apply_drop_nones)



from collections.abc import Iterable, Iterator, Mapping
from pprint import pformat
from typing import Any, NamedTuple

import cytoolz as cz


class Peeked[T](NamedTuple):
    value: T | tuple[T, ...]
    sequence: Iterator[T]


def peekn[T](data: Iterable[T], n: int) -> Iterator[T]:
    peeked = Peeked(*cz.itertoolz.peekn(n, data))
    print(f"Peeked {n} values: {peeked.value}")
    return peeked.sequence


def peek[T](data: Iterable[T]) -> Iterator[T]:
    peeked = Peeked(*cz.itertoolz.peek(data))
    print(f"Peeked value: {peeked.value}")
    return peeked.sequence


def dict_repr(
    v: Mapping[Any, Any],
    max_items: int = 20,
    depth: int = 3,
    width: int = 80,
    *,
    compact: bool = True,
) -> str:
    truncated = dict(list(v.items())[:max_items])
    suffix = "..." if len(v) > max_items else ""
    return pformat(truncated, depth=depth, width=width, compact=compact) + suffix



from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from typing import TYPE_CHECKING, Any

import cytoolz as cz

from .._core import IterWrapper

if TYPE_CHECKING:
    from .._dict import Dict


class BaseDict[T](IterWrapper[T]):
    def with_keys[K](self, keys: Iterable[K]) -> Dict[K, T]:
        """Create a Dict by zipping the iterable with keys.

        Args:
            keys (Iterable[K]): Iterable of keys to pair with the values.

        Returns:
            Dict[K, T]: Dict with the provided keys and iterable values.

        Example:
        ```python
        >>> import pyochain as pc
        >>> keys = ["a", "b", "c"]
        >>> values = [1, 2, 3]
        >>> pc.Iter.from_(values).with_keys(keys).inner()
        {'a': 1, 'b': 2, 'c': 3}
        >>> # This is equivalent to:
        >>> pc.Iter.from_(keys).zip(values).pipe(
        ...     lambda x: pc.Dict(x.into(dict)).inner()
        ... )
        {'a': 1, 'b': 2, 'c': 3}

        ```
        """
        from .._dict import Dict

        def _with_keys(data: Iterable[T]) -> Dict[K, T]:
            return Dict(dict(zip(keys, data, strict=False)))

        return self.into(_with_keys)

    def with_values[V](self, values: Iterable[V]) -> Dict[T, V]:
        """Create a Dict by zipping the iterable with values.

        Args:
            values (Iterable[V]): Iterable of values to pair with the keys.

        Returns:
            Dict[T, V]: Dict with the iterable as keys and provided values.

        Example:
        ```python
        >>> import pyochain as pc
        >>> keys = [1, 2, 3]
        >>> values = ["a", "b", "c"]
        >>> pc.Iter.from_(keys).with_values(values).inner()
        {1: 'a', 2: 'b', 3: 'c'}
        >>> # This is equivalent to:
        >>> pc.Iter.from_(keys).zip(values).pipe(
        ...     lambda x: pc.Dict(x.into(dict)).inner()
        ... )
        {1: 'a', 2: 'b', 3: 'c'}

        ```
        """
        from .._dict import Dict

        def _with_values(data: Iterable[T]) -> Dict[T, V]:
            return Dict(dict(zip(data, values, strict=False)))

        return self.into(_with_values)

    def reduce_by[K](
        self,
        key: Callable[[T], K],
        binop: Callable[[T, T], T],
    ) -> Dict[K, T]:
        """Perform a simultaneous groupby and reduction.

        Args:
            key (Callable[[T], K]): Function to compute the key for grouping.
            binop (Callable[[T, T], T]): Binary operation to reduce the grouped elements.

        Returns:
            Dict[K, T]: Dict with grouped and reduced elements.

        Example:
        ```python
        >>> from collections.abc import Iterable
        >>> import pyochain as pc
        >>> from operator import add, mul
        >>>
        >>> def is_even(x: int) -> bool:
        ...     return x % 2 == 0
        >>>
        >>> def group_reduce(data: Iterable[int]) -> int:
        ...     return pc.Iter.from_(data).reduce(add)
        >>>
        >>> data = pc.Seq([1, 2, 3, 4, 5])
        >>> data.iter().reduce_by(is_even, add).inner()
        {False: 9, True: 6}
        >>> data.iter().group_by(is_even).map_values(group_reduce).inner()
        {False: 9, True: 6}

        ```
        But the former does not build the intermediate groups, allowing it to operate in much less space.

        This makes it suitable for larger datasets that do not fit comfortably in memory

        Simple Examples:
        ```python
        >>> pc.Iter.from_([1, 2, 3, 4, 5]).reduce_by(is_even, add).inner()
        {False: 9, True: 6}
        >>> pc.Iter.from_([1, 2, 3, 4, 5]).reduce_by(is_even, mul).inner()
        {False: 15, True: 8}

        ```
        """
        from .._dict import Dict

        def _reduce_by(data: Iterable[T]) -> Dict[K, T]:
            return Dict(cz.itertoolz.reduceby(key, binop, data))

        return self.into(_reduce_by)

    def group_by[K](self, on: Callable[[T], K]) -> Dict[K, list[T]]:
        """Group elements by key function and return a Dict result.

        Args:
            on (Callable[[T], K]): Function to compute the key for grouping.

        Returns:
            Dict[K, list[T]]: Dict with grouped elements as lists.

        Example:
        ```python
        >>> import pyochain as pc
        >>> names = [
        ...     "Alice",
        ...     "Bob",
        ...     "Charlie",
        ...     "Dan",
        ...     "Edith",
        ...     "Frank",
        ... ]
        >>> pc.Iter.from_(names).group_by(len).sort()
        ... # doctest: +NORMALIZE_WHITESPACE
        {3: ['Bob', 'Dan'], 5: ['Alice', 'Edith', 'Frank'], 7: ['Charlie']}
        >>>
        >>> iseven = lambda x: x % 2 == 0
        >>> pc.Iter.from_([1, 2, 3, 4, 5, 6, 7, 8]).group_by(iseven)
        ... # doctest: +NORMALIZE_WHITESPACE
        {False: [1, 3, 5, 7], True: [2, 4, 6, 8]}

        ```
        Non-callable keys imply grouping on a member.
        ```python
        >>> data = [
        ...     {"name": "Alice", "gender": "F"},
        ...     {"name": "Bob", "gender": "M"},
        ...     {"name": "Charlie", "gender": "M"},
        ... ]
        >>> pc.Iter.from_(data).group_by("gender").sort()
        ... # doctest: +NORMALIZE_WHITESPACE
        {'F': [{'gender': 'F', 'name': 'Alice'}],
        'M': [{'gender': 'M', 'name': 'Bob'}, {'gender': 'M', 'name': 'Charlie'}]}

        ```
        """
        from .._dict import Dict

        def _group_by(data: Iterable[T]) -> Dict[K, list[T]]:
            return Dict(cz.itertoolz.groupby(on, data))

        return self.into(_group_by)

    def frequencies(self) -> Dict[T, int]:
        """Find number of occurrences of each value in the iterable.

        Returns:
            Dict[T, int]: Dict with element frequencies as counts.

        ```python
        >>> import pyochain as pc
        >>> data = ["cat", "cat", "ox", "pig", "pig", "cat"]
        >>> pc.Iter.from_(data).frequencies().inner()
        {'cat': 3, 'ox': 1, 'pig': 2}

        ```
        """
        from .._dict import Dict

        def _frequencies(data: Iterable[T]) -> Dict[T, int]:
            return Dict(cz.itertoolz.frequencies(data))

        return self.into(_frequencies)

    def count_by[K](self, key: Callable[[T], K]) -> Dict[K, int]:
        """Count elements of a collection by a key function.

        Args:
            key (Callable[[T], K]): Function to compute the key for counting.

        Returns:
            Dict[K, int]: Dict with count of elements for each key.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(["cat", "mouse", "dog"]).count_by(len).inner()
        {3: 2, 5: 1}
        >>> def iseven(x):
        ...     return x % 2 == 0
        >>> pc.Iter.from_([1, 2, 3]).count_by(iseven).inner()
        {False: 2, True: 1}

        ```
        """
        from .._dict import Dict

        def _count_by(data: Iterable[T]) -> Dict[K, int]:
            return Dict(cz.recipes.countby(key, data))

        return self.into(_count_by)

    def to_records[U: Sequence[Any]](self: BaseDict[U]) -> Dict[Any, Any]:
        """Transform an iterable of nested sequences into a nested dictionary.

        - Each inner sequence represents a path to a value in the dictionary.
        - The last element of each sequence is treated as the value
        - All preceding elements are treated as keys leading to that value.

        Returns:
            Dict[Any, Any]: Nested dictionary constructed from the sequences.

        Example:
        ```python
        >>> import pyochain as pc
        >>> arrays = [["a", "b", 1], ["a", "c", 2], ["d", 3]]
        >>> pc.Seq(arrays).to_records().inner()
        {'a': {'b': 1, 'c': 2}, 'd': 3}

        ```
        """
        from .._dict import Dict

        def _from_nested(
            arrays: Iterable[Sequence[Any]],
            parent: dict[Any, Any] | None = None,
        ) -> dict[Any, Any]:
            d: dict[Any, Any] = parent or {}
            for arr in arrays:
                if len(arr) > 1:
                    head, *tail = arr
                    if len(tail) == 1:
                        d[head] = tail[0]
                    else:
                        d[head] = _from_nested([tail], d.get(head, {}))
            return d

        return Dict(self.into(_from_nested))



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import TYPE_CHECKING, TypeIs

from .._core import Pipeable

if TYPE_CHECKING:
    from ._result import Result
    from ._states import NoneOption, Some


class OptionUnwrapError(RuntimeError): ...


class Option[T](Pipeable, ABC):
    """Type `Option[T]` represents an optional value.

    Every Option is either:
    - `Some` and contains a value
    - `None`, and does not.

    This is a common type in Rust, and is used to represent values that may be absent.

    In python, this is best tought of a an union type `T | None`,
    but with additional methods to operate on the contained value in a functional style.

    `Option[T]` and/or `T | None` types are very useful, as they have a number of uses:

        - Initial values
        - Union types
        - Return value where None is returned on error
        - Optional class fields
        - Optional function arguments

    The fact that `T | None` is a very common pattern in python,
    but without a dedicated structure/handling, leads to:
        - a lot of boilerplate code
        - potential bugs (even with type checkers)
        - less readable code (where does the None come from? is it expected?).

    `Option[T]` instances are commonly paired with pattern matching.
    This allow to query the presence of a value and take action, always accounting for the None case.
    """

    @abstractmethod
    def is_some(self) -> TypeIs[Some[T]]:  # type: ignore[misc]
        """Returns `True` if the option is a `Some` value.

        Uses `TypeIs[Some[T]]` for more precise type narrowing.

        Returns:
            TypeIs[Some[T]]: `True` if the option is a `Some` variant, `False` otherwise.

        Example:
        ```python
        >>> import pyochain as pc
        >>> x: Option[int] = pc.Some(2)
        >>> x.is_some()
        True
        >>> y: Option[int] = pc.NONE
        >>> y.is_some()
        False

        ```

        """
        ...

    @abstractmethod
    def is_none(self) -> TypeIs[NoneOption]:  # type: ignore[misc]
        """Returns `True` if the option is a `None` value.

        Uses `TypeIs[_None]` for more precise type narrowing.

        Returns:
            TypeIs[NoneOption]: `True` if the option is a `_None` variant, `False` otherwise.

        Example:
        ```python
        >>> import pyochain as pc
        >>> x: Option[int] = pc.Some(2)
        >>> x.is_none()
        False
        >>> y: Option[int] = pc.NONE
        >>> y.is_none()
        True

        ```

        """
        ...

    @abstractmethod
    def unwrap(self) -> T:
        """Returns the contained `Some` value.

        Returns:
            T: The contained `Some` value.

        Raises:
            OptionUnwrapError: If the option is `None`.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Some("car").unwrap()
        'car'

        ```
        ```python
        >>> import pyochain as pc
        >>> pc.NONE.unwrap()
        Traceback (most recent call last):
            ...
        pyochain._results._option.OptionUnwrapError: called `unwrap` on a `None`

        ```

        """
        ...

    def expect(self, msg: str) -> T:
        """Returns the contained `Some` value.

        Raises an exception with a provided message if the value is `None`.

        Args:
            msg (str): The message to include in the exception if the result is `None`.

        Returns:
            T: The contained `Some` value.

        Raises:
            OptionUnwrapError: If the result is `None`.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Some("value").expect("fruits are healthy")
        'value'
        >>> pc.NONE.expect("fruits are healthy")
        Traceback (most recent call last):
            ...
        pyochain._results._option.OptionUnwrapError: fruits are healthy (called `expect` on a `None`)

        ```

        """
        if self.is_some():
            return self.unwrap()
        msg = f"{msg} (called `expect` on a `None`)"
        raise OptionUnwrapError(msg)

    def unwrap_or(self, default: T) -> T:
        """Returns the contained `Some` value or a provided default.

        Args:
            default (T): The value to return if the result is `None`.

        Returns:
            T: The contained `Some` value or the provided default.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Some("car").unwrap_or("bike")
        'car'
        >>> pc.NONE.unwrap_or("bike")
        'bike'

        ```

        """
        return self.unwrap() if self.is_some() else default

    def unwrap_or_else(self, f: Callable[[], T]) -> T:
        """Returns the contained `Some` value or computes it from a function.

        Args:
            f (Callable[[], T]): A function that returns a default value if the result is `None`.

        Returns:
            T: The contained `Some` value or the result of the function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> k = 10
        >>> pc.Some(4).unwrap_or_else(lambda: 2 * k)
        4
        >>> pc.NONE.unwrap_or_else(lambda: 2 * k)
        20

        ```

        """
        return self.unwrap() if self.is_some() else f()

    def map[U](self, f: Callable[[T], U]) -> Option[U]:
        """Maps an `Option[T]` to `Option[U]`.

        Done by applying a function to a contained `Some` value,
        leaving a `None` value untouched.

        Args:
            f (Callable[[T], U]): The function to apply to the `Some` value.

        Returns:
            Option[U]: A new `Option` with the mapped value if `Some`, otherwise `None`.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Some("Hello, World!").map(len)
        Some(value=13)
        >>> pc.NONE.map(len)
        NONE

        ```

        """
        from ._states import NONE, Some

        return Some(f(self.unwrap())) if self.is_some() else NONE

    def and_then[U](self, f: Callable[[T], Option[U]]) -> Option[U]:
        """Calls a function if the option is `Some`, otherwise returns `None`.

        Args:
            f (Callable[[T], Option[U]]): The function to call with the `Some` value.

        Returns:
            Option[U]: The result of the function if `Some`, otherwise `None`.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def sq(x: int) -> Option[int]:
        ...     return pc.Some(x * x)
        >>> def nope(x: int) -> Option[int]:
        ...     return pc.NONE
        >>> pc.Some(2).and_then(sq).and_then(sq)
        Some(value=16)
        >>> pc.Some(2).and_then(sq).and_then(nope)
        NONE
        >>> pc.Some(2).and_then(nope).and_then(sq)
        NONE
        >>> pc.NONE.and_then(sq).and_then(sq)
        NONE

        ```

        """
        from ._states import NONE

        return f(self.unwrap()) if self.is_some() else NONE

    def or_else(self, f: Callable[[], Option[T]]) -> Option[T]:
        """Returns the `Option[T]` if it contains a value, otherwise calls a function and returns the result.

        Args:
            f (Callable[[], Option[T]]): The function to call if the option is `None`.

        Returns:
            Option[T]: The original `Option` if it is `Some`, otherwise the result of the function.

        Example:
        ```python
        >>> import pyochain as pc
        >>> def nobody() -> Option[str]:
        ...     return pc.NONE
        >>> def vikings() -> Option[str]:
        ...     return pc.Some("vikings")
        >>> pc.Some("barbarians").or_else(vikings)
        Some(value='barbarians')
        >>> pc.NONE.or_else(vikings)
        Some(value='vikings')
        >>> pc.NONE.or_else(nobody)
        NONE

        ```

        """
        return self if self.is_some() else f()

    def zip[U](self, other: Option[U]) -> Option[tuple[T, U]]:
        """Returns an `Option[tuple[T, U]]` containing a tuple of the values if both options are `Some`, otherwise returns `NONE`.

        Args:
            other (Option[U]): The other option to zip with.

        Returns:
            Option[tuple[T, U]]: Some((self, other)) if both are Some, otherwise NONE.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Some(1).zip(pc.Some('a'))
        Some(value=(1, 'a'))
        >>> pc.Some(1).zip(pc.NONE)
        NONE
        >>> pc.NONE.zip(pc.Some('a'))
        NONE

        ```

        """
        from ._states import NONE, Some

        if self.is_some() and other.is_some():
            return Some((self.unwrap(), other.unwrap()))
        return NONE

    def zip_with[U, R](self, other: Option[U], f: Callable[[T, U], R]) -> Option[R]:
        """Zips `self` and another `Option` with function `f`.

        If `self` is `Some(s)` and other is `Some(o)`, this method returns `Some(f(s, o))`.

        Otherwise, `NONE` is returned.

        Args:
            other (Option[U]): The second option.
            f (Callable[[T, U], R]): The function to apply to the unwrapped values.

        Returns:
            Option[R]: The resulting option after applying the function.

        Examples:
        ```python
        >>> from dataclasses import dataclass
        >>> import pyochain as pc
        >>>
        >>> @dataclass
        ... class Point:
        ...     x: float
        ...     y: float
        >>>
        >>> x = pc.Some(17.5)
        >>> y = pc.Some(42.7)
        >>> x.zip_with(y, Point)
        Some(value=Point(x=17.5, y=42.7))
        >>> x.zip_with(pc.NONE, Point)
        NONE

        ```

        """
        from ._states import NONE, Some

        if self.is_some() and other.is_some():
            return Some(f(self.unwrap(), other.unwrap()))
        return NONE

    def ok_or[E](self, err: E) -> Result[T, E]:
        """Converts the option to a `Result`.

        Args:
            err (E): The error value to use if the option is `NONE`.

        Returns:
            Result[T, E]: `Ok(v)` if `Some(v)`, otherwise `Err(err)`.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Some(1).ok_or('fail')
        Ok(value=1)
        >>> pc.NONE.ok_or('fail')
        Err(error='fail')

        ```

        """
        from ._states import Err, Ok

        return Ok(self.unwrap()) if self.is_some() else Err(err)

    def ok_or_else[E](self, err: Callable[[], E]) -> Result[T, E]:
        """Converts the option to a Result.

        Args:
            err (Callable[[], E]): A function returning the error value if the option is NONE.

        Returns:
            Result[T, E]: Ok(v) if Some(v), otherwise Err(err()).

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Some(1).ok_or_else(lambda: 'fail')
        Ok(value=1)
        >>> pc.NONE.ok_or_else(lambda: 'fail')
        Err(error='fail')

        ```

        """
        from ._states import Err, Ok

        return Ok(self.unwrap()) if self.is_some() else Err(err())

    def map_or[U](self, f: Callable[[T], U], default: U) -> U:
        """Returns the result of applying a function to the contained value if Some, otherwise returns the default value.

        Args:
            f (Callable[[T], U]): The function to apply to the contained value.
            default (U): The default value to return if NONE.

        Returns:
            U: The result of f(self.unwrap()) if Some, otherwise default.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Some(2).map_or(lambda x: x * 10, 0)
        20
        >>> pc.NONE.map_or(lambda x: x * 10, 0)
        0

        ```

        """
        return f(self.unwrap()) if self.is_some() else default

    def map_or_else[U](self, f: Callable[[T], U], default: Callable[[], U]) -> U:
        """Returns the result of applying a function to the contained value if Some, otherwise computes a default value.

        Args:
            f (Callable[[T], U]): The function to apply to the contained value.
            default (Callable[[], U]): A function returning the default value if NONE.

        Returns:
            U: The result of f(self.unwrap()) if Some, otherwise default().

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Some(2).map_or_else(lambda x: x * 10, lambda: 0)
        20
        >>> pc.NONE.map_or_else(lambda x: x * 10, lambda: 0)
        0

        ```

        """
        return f(self.unwrap()) if self.is_some() else default()



"""pyochain - A functional programming library for Python."""

from ._dict import Dict
from ._iter import Iter, Seq
from ._results import NONE, Err, Ok, Option, Result, ResultUnwrapError, Some, Wrapper

__all__ = [
    "NONE",
    "Dict",
    "Err",
    "Iter",
    "Ok",
    "Option",
    "Result",
    "ResultUnwrapError",
    "Seq",
    "Some",
    "Wrapper",
]
